"""
This module provides a function to compute thickness

"""

import logging
import os

import nibabel as nb
import numpy as np
import pandas as pd
from lapy import TetMesh, TriaMesh, io
from scipy.sparse import csgraph as sc

from .utils.get_levelsets import levelsetsTetra

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS

# ------------------------------------------------------------------------------
# MAIN FUNCTION
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# computeThickness


def computeThickness(params):
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Computing thickness")
    print()

    # get data

    IN_MESH = os.path.join(params.OUTDIR, "tetra-cube", params.HEMI + ".seam.rm.cut.vtk")
    IN_FUNC = os.path.join(params.OUTDIR, "tetra-cube", params.HEMI + ".uvw.seam.rm.cut.vtk")
    OUT_DIR = os.path.join(params.OUTDIR, "thickness")

    # load mesh

    tetMesh = TetMesh.read_vtk(IN_MESH)

    v4 = tetMesh.v
    t4 = tetMesh.t

    # load function

    tetMeshUVW = TetMesh.read_vtk(IN_FUNC)

    p4 = tetMeshUVW.v
    q4 = tetMeshUVW.t

    # determine levels

    lLVL4x = np.linspace(params.internal.THXn, params.internal.THXp, params.internal.THXk)
    lLVL4y = np.linspace(params.internal.THYn, params.internal.THYp, params.internal.THYk)
    lLVL4z = np.linspace(params.internal.THZn, params.internal.THZp, params.internal.THZk)

    # compute levelsets in original space

    vLVL4x, tLVL4x, iLVL4x, jLVL4x, oLVL4x = levelsetsTetra(v4, t4, p4[:, 0], lLVL4x)
    vLVL4y, tLVL4y, iLVL4y, jLVL4y, oLVL4y = levelsetsTetra(v4, t4, p4[:, 1], lLVL4y)
    vLVL4z, tLVL4z, iLVL4z, jLVL4z, oLVL4z = levelsetsTetra(v4, t4, p4[:, 2], lLVL4z)

    # compute levelsets in parameter space

    vLVL4px, tLVL4px, iLVL4px, jLVL4px, oLVL4px = levelsetsTetra(p4, t4, p4[:, 0], lLVL4x)
    vLVL4py, tLVL4py, iLVL4py, jLVL4py, oLVL4py = levelsetsTetra(p4, t4, p4[:, 1], lLVL4y)
    vLVL4pz, tLVL4pz, iLVL4pz, jLVL4pz, oLVL4pz = levelsetsTetra(p4, t4, p4[:, 2], lLVL4z)

    # write out single slices in x, y, z directions

    if params.internal.CLEANUP is False:
        for i in range(0, len(vLVL4px)):
            TriaMesh.write_vtk(
                TriaMesh(vLVL4x[i], tLVL4x[i] - 1), os.path.join(OUT_DIR, params.HEMI + ".vlx.lvl" + str(i) + ".vtk")
            )
            TriaMesh.write_vtk(
                TriaMesh(vLVL4px[i], tLVL4px[i] - 1),
                os.path.join(OUT_DIR, params.HEMI + ".uvw.vlx.lvl" + str(i) + ".vtk"),
            )

        for i in range(0, len(vLVL4py)):
            TriaMesh.write_vtk(
                TriaMesh(vLVL4y[i], tLVL4y[i] - 1), os.path.join(OUT_DIR, params.HEMI + ".vly.lvl" + str(i) + ".vtk")
            )
            TriaMesh.write_vtk(
                TriaMesh(vLVL4py[i], tLVL4py[i] - 1),
                os.path.join(OUT_DIR, params.HEMI + ".uvw.vly.lvl" + str(i) + ".vtk"),
            )

        for i in range(0, len(vLVL4pz)):
            TriaMesh.write_vtk(
                TriaMesh(vLVL4z[i], tLVL4z[i] - 1), os.path.join(OUT_DIR, params.HEMI + ".vlz.lvl" + str(i) + ".vtk")
            )
            TriaMesh.write_vtk(
                TriaMesh(vLVL4pz[i], tLVL4pz[i] - 1),
                os.path.join(OUT_DIR, params.HEMI + ".uvw.vlz.lvl" + str(i) + ".vtk"),
            )

    # now loop across levelsets

    origV4 = np.empty((len(vLVL4px), len(vLVL4py), len(vLVL4pz), 3))
    origV4[:] = np.nan

    origV4flat = list()

    ctr = np.empty((len(vLVL4px), len(vLVL4py), len(vLVL4pz)))
    ctr[:] = np.nan

    ctrflat = list()

    msgNoCmTet = 0
    msgCmTetNoInt = 0
    msgCmTetMnInt = 0

    for lx in range(0, len(vLVL4px)):
        for ly in range(0, len(vLVL4py)):
            for lz in range(0, len(vLVL4pz)):
                # which tetras are present in all three functions? note that
                # this can be more than one tetra, and we do not know the
                # number in advance.

                # variant based on parameter levelsets:
                tmp = np.intersect1d(iLVL4px[lx], np.intersect1d(iLVL4py[ly], iLVL4pz[lz]))

                if len(tmp) == 0:
                    msgNoCmTet += 1

                    origV4[lx, ly, lz, :] = [np.nan, np.nan, np.nan]
                    ctr[lx, ly, lz] = np.nan

                else:
                    # if using p4, which tetra holds point p4(lx,ly,lz)?

                    # create a barycentric coordinate system with the first
                    # vertex as the basis; point is inside the tetra if all
                    # barycentric oordinates are within the range [0,1]
                    # (sign matters!)

                    # then map from barycentric ctr back to xyz in v4 space

                    # Tp is the transformation matrix for the parameter space
                    # Tv is the transformation matrix for the vertex space

                    # b is the vector of barycentric coordinates

                    # ctr is the tetra within which the point is located

                    Tp = np.empty((3, 3, len(tmp)))
                    Tp[:] = np.nan

                    b = np.empty((3, len(tmp)))
                    b[:] = np.nan

                    for i in range(0, len(tmp)):
                        Tp[:, :, i] = np.array(
                            [
                                p4[t4[tmp[i], 1], :] - p4[t4[tmp[i], 0], :],
                                p4[t4[tmp[i], 2], :] - p4[t4[tmp[i], 0], :],
                                p4[t4[tmp[i], 3], :] - p4[t4[tmp[i], 0], :],
                            ]
                        ).transpose()

                        if np.linalg.matrix_rank(Tp[:, :, i]) == 3:
                            b[:, i] = np.matmul(
                                np.linalg.inv(Tp[:, :, i]),
                                np.array([[lLVL4x[lx], lLVL4y[ly], lLVL4z[lz]] - p4[t4[tmp[i], 0], :]]).transpose(),
                            ).squeeze()

                        else:
                            b[:, i] = np.nan

                    b = np.concatenate((b, (1 - np.sum(b, axis=0)).reshape((1, b.shape[1]))), axis=0)

                    tmpi = np.where(np.sum((b >= 0) & (b <= 1), axis=0) == 4)[0]

                    if len(tmpi) == 0:
                        msgCmTetNoInt += 1
                    elif len(tmpi) > 1:
                        msgCmTetMnInt += 1
                    else:
                        Tv = np.array(
                            [
                                v4[t4[tmp[tmpi[0]], 1], :] - v4[t4[tmp[tmpi[0]], 0], :],
                                v4[t4[tmp[tmpi[0]], 2], :] - v4[t4[tmp[tmpi[0]], 0], :],
                                v4[t4[tmp[tmpi[0]], 3], :] - v4[t4[tmp[tmpi[0]], 0], :],
                            ]
                        ).transpose()

                        origV4[lx, ly, lz, :] = np.matmul(Tv, b[0:3, tmpi[0]]) + v4[t4[tmp[tmpi[0]], 0], :]

                        ctr[lx, ly, lz] = tmp[tmpi[0]]

                if any(np.isnan(origV4[lx, ly, lz, :])) is False:
                    origV4flat.append(np.hstack((np.array([lx, ly, lz]), origV4[lx, ly, lz, :])))

                if np.isnan(ctr[lx, ly, lz]) is False:
                    ctrflat.append(np.hstack((np.array([lx, ly, lz]), ctr[lx, ly, lz])))

    origV4flat = np.array(origV4flat)
    ctrflat = np.array(ctr.flat)

    LOGGER.info("No common tetra found: " + str(msgNoCmTet))
    LOGGER.info("Common tetra, but no intersection: " + str(msgCmTetNoInt))
    LOGGER.info("Common tetra, but too many intersections: " + str(msgCmTetMnInt))

    # -------------------------------------------------------------------------
    # lines

    # x->yz
    llx = [[list() for k in range(np.shape(origV4)[2])] for z in range(np.shape(origV4)[1])]
    for iy in range(0, np.shape(origV4)[1]):
        for iz in range(0, np.shape(origV4)[2]):
            llx[iy][iz] = origV4flat[np.logical_and(origV4flat[:, 1] == iy, origV4flat[:, 2] == iz), 0:6]

    # y->xz
    lly = [[list() for k in range(np.shape(origV4)[2])] for z in range(np.shape(origV4)[0])]
    for ix in range(0, np.shape(origV4)[0]):
        for iz in range(0, np.shape(origV4)[2]):
            lly[ix][iz] = origV4flat[np.logical_and(origV4flat[:, 0] == ix, origV4flat[:, 2] == iz), 0:6]

    # z->xy
    llz = [[list() for k in range(np.shape(origV4)[1])] for z in range(np.shape(origV4)[0])]
    for ix in range(0, np.shape(origV4)[0]):
        for iy in range(0, np.shape(origV4)[1]):
            llz[ix][iy] = origV4flat[np.logical_and(origV4flat[:, 0] == ix, origV4flat[:, 1] == iy), 0:6]

    # --------------------------------------------------------------------------
    # create fake streamlines (will also work with non-continuous lines)

    # x

    vlx = list()
    tlx = list()
    nlx = list()
    llxLgth = np.full((np.shape(origV4)[1], np.shape(origV4)[2]), np.nan)

    for ii in range(len(llx)):
        for ij in range(len(llx[ii])):
            if len(llx[ii][ij]) > 1:
                tmp = list()

                for ik in range(len(llx[ii][ij]) - 1):
                    if llx[ii][ij][ik, 0] + 1 == llx[ii][ij][ik + 1, 0]:
                        vlx.append(llx[ii][ij][ik, 3:6])

                        tlx.append([len(vlx), len(vlx), len(vlx) + 1])

                        nlx.append([ii, ij, ik, np.linalg.norm(llx[ii][ij][ik + 1, 3:6] - llx[ii][ij][ik, 3:6])])

                        tmp.append(np.linalg.norm(llx[ii][ij][ik + 1, 3:6] - llx[ii][ij][ik, 3:6]))

                    elif llx[ii][ij][ik - 1, 0] + 1 == llx[ii][ij][ik, 0]:
                        # close a potentially open previous line
                        vlx.append(llx[ii][ij][ik, 3:6])

                if llx[ii][ij][ik + 1, 0] - 1 == llx[ii][ij][ik, 0]:
                    vlx.append(llx[ii][ij][ik + 1, 3:6])

                if np.sum(tmp):  # empty lists are false
                    llxLgth[ii, ij] = np.sum(tmp)

    TriaMesh.write_vtk(
        TriaMesh(np.array(vlx), np.asarray(tlx) - 1), os.path.join(OUT_DIR, params.HEMI + ".grid-lines-x.vtk")
    )

    dfx = pd.DataFrame(llxLgth)

    dz = zip(range(0, dfx.shape[1]), ["z" + str(z) for z in range(0, dfx.shape[1])])
    dfx = dfx.rename(columns=dict(dz))

    dy = zip(range(0, dfx.shape[0]), ["y" + str(z) for z in range(0, dfx.shape[0])])
    dfx = dfx.rename(index=dict(dy))

    dfx.to_csv(os.path.join(OUT_DIR, params.HEMI + ".grid-segments-x.csv"))

    # y

    vly = list()
    tly = list()
    nly = list()
    llyLgth = np.full((np.shape(origV4)[0], np.shape(origV4)[2]), np.nan)

    for ii in range(len(lly)):
        for ij in range(len(lly[ii])):
            if len(lly[ii][ij]) > 1:
                tmp = list()

                for ik in range(len(lly[ii][ij]) - 1):
                    if lly[ii][ij][ik, 1] + 1 == lly[ii][ij][ik + 1, 1]:
                        vly.append(lly[ii][ij][ik, 3:6])

                        tly.append([len(vly), len(vly), len(vly) + 1])

                        nly.append([ii, ij, ik, np.linalg.norm(lly[ii][ij][ik + 1, 3:6] - lly[ii][ij][ik, 3:6])])

                        tmp.append(np.linalg.norm(lly[ii][ij][ik + 1, 3:6] - lly[ii][ij][ik, 3:6]))

                    elif lly[ii][ij][ik - 1, 1] + 1 == lly[ii][ij][ik, 1]:
                        # close a potentially open previous line
                        vly.append(lly[ii][ij][ik, 3:6])

                if lly[ii][ij][ik + 1, 1] - 1 == lly[ii][ij][ik, 1]:
                    vly.append(lly[ii][ij][ik + 1, 3:6])

                if np.sum(tmp):  # empty lists are false
                    llyLgth[ii, ij] = np.sum(tmp)

    TriaMesh.write_vtk(
        TriaMesh(np.array(vly), np.asarray(tly) - 1), os.path.join(OUT_DIR, params.HEMI + ".grid-lines-y.vtk")
    )

    dfy = pd.DataFrame(llyLgth)

    dz = zip(range(0, dfy.shape[1]), ["z" + str(z) for z in range(0, dfy.shape[1])])
    dfy = dfy.rename(columns=dict(dz))

    dx = zip(range(0, dfy.shape[0]), ["x" + str(z) for z in range(0, dfy.shape[0])])
    dfy = dfy.rename(index=dict(dx))

    dfy.to_csv(os.path.join(OUT_DIR, params.HEMI + ".grid-segments-y.csv"))

    # z

    vlz = list()
    tlz = list()
    nlz = list()
    llzLgth = np.full((np.shape(origV4)[0], np.shape(origV4)[1]), np.nan)

    for ii in range(len(llz)):
        for ij in range(len(llz[ii])):
            if len(llz[ii][ij]) > 1:
                tmp = list()

                for ik in range(len(llz[ii][ij]) - 1):
                    if llz[ii][ij][ik, 2] + 1 == llz[ii][ij][ik + 1, 2]:
                        vlz.append(llz[ii][ij][ik, 3:6])

                        tlz.append([len(vlz), len(vlz), len(vlz) + 1])

                        nlz.append([ii, ij, ik, np.linalg.norm(llz[ii][ij][ik + 1, 3:6] - llz[ii][ij][ik, 3:6])])

                        tmp.append(np.linalg.norm(llz[ii][ij][ik + 1, 3:6] - llz[ii][ij][ik, 3:6]))

                    elif llz[ii][ij][ik - 1, 2] + 1 == llz[ii][ij][ik, 2]:
                        # close a potentially open previous line
                        vlz.append(llz[ii][ij][ik, 3:6])

                if llz[ii][ij][ik + 1, 2] - 1 == llz[ii][ij][ik, 2]:
                    vlz.append(llz[ii][ij][ik + 1, 3:6])

                if np.sum(tmp):  # empty lists are false
                    llzLgth[ii, ij] = np.sum(tmp)

    TriaMesh.write_vtk(
        TriaMesh(np.array(vlz), np.asarray(tlz) - 1), os.path.join(OUT_DIR, params.HEMI + ".grid-lines-z.vtk")
    )

    dfz = pd.DataFrame(llzLgth)

    dy = zip(range(0, dfz.shape[1]), ["y" + str(z) for z in range(0, dfz.shape[1])])
    dfz = dfz.rename(columns=dict(dy))

    dx = zip(range(0, dfz.shape[0]), ["x" + str(z) for z in range(0, dfz.shape[0])])
    dfz = dfz.rename(index=dict(dx))

    dfz.to_csv(os.path.join(OUT_DIR, params.HEMI + ".grid-segments-z.csv"))

    # --------------------------------------------------------------------------
    # export origV4flat

    pd.DataFrame(origV4flat).to_csv(os.path.join(OUT_DIR, params.HEMI + ".grid-lines.csv"), header=False, index=False)

    # --------------------------------------------------------------------------
    # create checkerboard (we only do it for one dimension: z)

    origV4flatTria = dict()
    origV4flatCol = dict()

    minIdx = 0
    midIdx = int(np.round((len(lLVL4z) - 1) / 2))
    maxIdx = len(lLVL4z) - 1

    # fmt: off
    for k in [minIdx, midIdx, maxIdx]:

        origV4flatTria[k] = list()
        origV4flatCol[k] = list()

        for i in range(0, len(llz) - 1):

            if params.internal.allow_ragged_surfaces or (
            (len(np.unique(origV4flat[np.logical_and(origV4flat[:, 0] == i, origV4flat[:, 2] == k), 1])) == params.internal.THYk) and
            (len(np.unique(origV4flat[np.logical_and(origV4flat[:, 0] == i+1, origV4flat[:, 2] == k), 1])) == params.internal.THYk)
            ):

                for j in range(0, len(llz[i])-1):

                    # result may be empty, but that's OK: nothing will be added then

                    # we are doing this per square, could be done per tria also (split
                    # into two if-clauses)

                    if params.internal.allow_ragged_trias is False:

                        if all((
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0])>0)
                            )):

                            origV4flatTria[k].append(
                                [
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0]
                                ]
                            )

                            origV4flatTria[k].append(
                                [
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0]
                                ]
                            )

                            origV4flatCol[k].append(np.mod(i + j, 2))
                            origV4flatCol[k].append(np.mod(i + j, 2))

                    elif params.internal.allow_ragged_trias is True:

                        if all((
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0])>0)
                            )):

                            origV4flatTria[k].append(
                                [
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0]
                                ]
                            )

                            origV4flatCol[k].append(np.mod(i + j, 2))

                        if all((
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0])>0),
                            (len(np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0])>0)
                            )):

                            origV4flatTria[k].append(
                                [
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j    , origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i    , np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0],
                                    np.where(np.logical_and(origV4flat[:, 0] == i + 1, np.logical_and(origV4flat[:, 1] == j + 1, origV4flat[:, 2] == k)))[0][0]
                                ]
                            )

                            origV4flatCol[k].append(np.mod(i + j, 2))

    # fmt: on

    #

    vMin = origV4flat[:, 3:6]
    vMid = origV4flat[:, 3:6]
    vMax = origV4flat[:, 3:6]

    tMin = np.array(origV4flatTria[minIdx])
    tMid = np.array(origV4flatTria[midIdx])
    tMax = np.array(origV4flatTria[maxIdx])

    #

    cMin = origV4flat[:, 0:3]
    cMid = origV4flat[:, 0:3]
    cMax = origV4flat[:, 0:3]

    triaMin = TriaMesh(vMin, tMin)
    triaMid = TriaMesh(vMid, tMid)
    triaMax = TriaMesh(vMax, tMax)

    vMinKeep, vMinDel = triaMin.rm_free_vertices_()
    vMidKeep, vMidDel = triaMid.rm_free_vertices_()
    vMaxKeep, vMaxDel = triaMax.rm_free_vertices_()

    cMinRm = cMid[vMinKeep, :]
    cMidRm = cMid[vMidKeep, :]
    cMaxRm = cMid[vMaxKeep, :]

    triaMinRm = TriaMesh(triaMin.v, triaMin.t)
    triaMidRm = TriaMesh(triaMid.v, triaMid.t)
    triaMaxRm = TriaMesh(triaMax.v, triaMax.t)

    #

    if params.internal.no_orient is False:
        if sc.connected_components(triaMinRm._construct_adj_sym())[0] == 1:
            triaMinRm.orient_()
            skipOrientMin = False
        else:
            LOGGER.warning(
                "Warning: exterior surface contains more than one component, not orienting. Check your results."
            )
            skipOrientMin = True
        if sc.connected_components(triaMidRm._construct_adj_sym())[0] == 1:
            triaMidRm.orient_()
            skipOrientMid = False
        else:
            LOGGER.warning("Warning: mid-surface contains more than one component, not orienting. Check your results.")
            skipOrientMid = True
        if sc.connected_components(triaMaxRm._construct_adj_sym())[0] == 1:
            triaMaxRm.orient_()
            skipOrientMax = False
        else:
            LOGGER.warning(
                "Warning: interior surface contains more than one component, not orienting. Check your results."
            )
            skipOrientMax = True
    else:
        skipOrientMin = True
        skipOrientMid = True
        skipOrientMax = True

    #

    TriaMesh.write_vtk(triaMinRm, os.path.join(OUT_DIR, params.HEMI + ".ext-surface.vtk"))
    TriaMesh.write_vtk(triaMidRm, os.path.join(OUT_DIR, params.HEMI + ".mid-surface.vtk"))
    TriaMesh.write_vtk(triaMaxRm, os.path.join(OUT_DIR, params.HEMI + ".int-surface.vtk"))

    pd.DataFrame(np.concatenate((cMinRm, triaMinRm.v), axis=1)).to_csv(
        os.path.join(OUT_DIR, params.HEMI + ".ext-surface.csv"), header=False, index=False
    )
    pd.DataFrame(np.concatenate((cMidRm, triaMidRm.v), axis=1)).to_csv(
        os.path.join(OUT_DIR, params.HEMI + ".mid-surface.csv"), header=False, index=False
    )
    pd.DataFrame(np.concatenate((cMaxRm, triaMaxRm.v), axis=1)).to_csv(
        os.path.join(OUT_DIR, params.HEMI + ".int-surface.csv"), header=False, index=False
    )

    # --------------------------------------------------------------------------
    # export thickness

    thickness = list()

    for i in range(0, len(cMidRm)):
        thickness.append(np.array(dfz)[int(cMidRm[i, 0]), int(cMidRm[i, 1])])

    io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".mid-surface.thickness.psol"), np.array(thickness))

    nb.freesurfer.save(
        nb.freesurfer.MGHImage(dataobj=np.array(thickness).astype("float32"), affine=None),
        filename=os.path.join(OUT_DIR, params.HEMI + ".mid-surface.thickness.mgh"),
    )

    # --------------------------------------------------------------------------
    # export curvature

    if skipOrientMin is False:
        u1Min, u2Min, c1Min, c2Min = triaMinRm.curvature_tria()

        mean_curvMin = (c1Min + c2Min) / 2

        gauss_curvMin = c1Min * c2Min

        mean_curvMin = triaMinRm.map_tfunc_to_vfunc(mean_curvMin)

        gauss_curvMin = triaMinRm.map_tfunc_to_vfunc(gauss_curvMin)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".ext-surface.mean-curv.psol"), mean_curvMin)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".ext-surface.gauss-curv.psol"), gauss_curvMin)

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=mean_curvMin.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".ext-surface.mean-curv.mgh"),
        )

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=gauss_curvMin.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".ext-surface.gauss-curv.mgh"),
        )

        pd.DataFrame(np.concatenate((cMinRm, np.array(mean_curvMin, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".ext-surface.mean-curv.csv"), header=False, index=False
        )

        pd.DataFrame(np.concatenate((cMinRm, np.array(gauss_curvMin, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".ext-surface.gauss-curv.csv"), header=False, index=False
        )

    else:
        LOGGER.warning("Warning: Not computing curvature for exterior surface, check surface.")

    if skipOrientMid is False:
        u1Mid, u2Mid, c1Mid, c2Mid = triaMidRm.curvature_tria()

        mean_curvMid = (c1Mid + c2Mid) / 2

        gauss_curvMid = c1Mid * c2Mid

        mean_curvMid = triaMidRm.map_tfunc_to_vfunc(mean_curvMid)

        gauss_curvMid = triaMidRm.map_tfunc_to_vfunc(gauss_curvMid)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".mid-surface.mean-curv.psol"), mean_curvMid)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".mid-surface.gauss-curv.psol"), gauss_curvMid)

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=mean_curvMid.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".mid-surface.mean-curv.mgh"),
        )

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=gauss_curvMid.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".mid-surface.gauss-curv.mgh"),
        )

        pd.DataFrame(np.concatenate((cMidRm, np.array(mean_curvMid, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".mid-surface.mean-curv.csv"), header=False, index=False
        )

        pd.DataFrame(np.concatenate((cMidRm, np.array(gauss_curvMid, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".mid-surface.gauss-curv.csv"), header=False, index=False
        )

    else:
        LOGGER.warning("Warning: Not computing curvature for mid-surface, check surface.")

    if skipOrientMax is False:
        u1Max, u2Max, c1Max, c2Max = triaMaxRm.curvature_tria()

        mean_curvMax = (c1Max + c2Max) / 2

        gauss_curvMax = c1Max * c2Max

        mean_curvMax = triaMaxRm.map_tfunc_to_vfunc(mean_curvMax)

        gauss_curvMax = triaMaxRm.map_tfunc_to_vfunc(gauss_curvMax)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".int-surface.mean-curv.psol"), mean_curvMax)

        io.write_vfunc(os.path.join(OUT_DIR, params.HEMI + ".int-surface.gauss-curv.psol"), gauss_curvMax)

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=mean_curvMax.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".int-surface.mean-curv.mgh"),
        )

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=gauss_curvMax.astype("float32"), affine=None),
            filename=os.path.join(OUT_DIR, params.HEMI + ".int-surface.gauss-curv.mgh"),
        )

        pd.DataFrame(np.concatenate((cMaxRm, np.array(mean_curvMax, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".int-surface.mean-curv.csv"), header=False, index=False
        )

        pd.DataFrame(np.concatenate((cMaxRm, np.array(gauss_curvMax, ndmin=2).transpose()), axis=1)).to_csv(
            os.path.join(OUT_DIR, params.HEMI + ".int-surface.gauss-curv.csv"), header=False, index=False
        )

    else:
        LOGGER.warning("Warning: Not computing curvature for interior surface, check surface.")

    # -----------------------------------------------------------------------------
    # create hull

    # here: fast --> slow: d1 --> d2 --> d3

    f = origV4flat

    d1 = params.internal.THXk
    d2 = params.internal.THYk
    d3 = params.internal.THZk

    #

    vert = np.full((d1, d2, d3, 3), np.nan)
    tria = list()

    # fmt: off
    for f_i in range(0, len(f)):

        # get indices
        i, j, k = f[f_i, 0:3].astype(int)

        # assign coordinates
        vert[i,j,k,:] = f[f_i, 3:6]

        # get trias

        # x==0|x==d1-1
        if (i==0) or i==(d1-1):

            # x:yz
            if (np.mod(j+1, d2)>0) and (np.mod(k+1, d3)>0):
                tria.append([i*d2*d3+j*d3+k, i*d2*d3+j*d3+k+1, i*d2*d3+j*d3+k+d3+1])
                tria.append([i*d2*d3+j*d3+k, i*d2*d3+j*d3+k+d3+1, i*d2*d3+j*d3+k+d3])

        # y==0|y==d2-1
        if (j==0) or (j==(d2-1)):

            # y:xz
            if (np.mod(i+1, d1)>0) and (np.mod(k+1, d3)>0):
                tria.append([i*d2*d3+j*d3+k, i*d2*d3+j*d3+k+d3*d2, i*d2*d3+j*d3+k+1])
                tria.append([i*d2*d3+j*d3+k+1, i*d2*d3+j*d3+k+d3*d2, i*d2*d3+j*d3+k+d3*d2+1])

        # z==0|z==d3-1
        if (k==0) or (k==(d3-1)):

            # z:xy
            if (np.mod(i+1, d1)>0) and (np.mod(j+1, d2)>0):
                tria.append([i*d2*d3+j*d3+k, i*d2*d3+j*d3+k+d2*d3, i*d2*d3+j*d3+k+d3])
                tria.append([i*d2*d3+j*d3+k+d3, i*d2*d3+j*d3+k+d2*d3, i*d2*d3+j*d3+k+d2*d3+d3])

    # fmt: on
    vertList = np.reshape(vert, (d1 * d2 * d3, 3))
    triaNoNan = np.array(tria)[np.sum(np.sum(np.isnan(vertList[tria, :]), axis=2), axis=1) == 0, :]

    TriaMesh.write_vtk(TriaMesh(vertList, triaNoNan), os.path.join(OUT_DIR, params.HEMI + ".hull.vtk"))

    # --------------------------------------------------------------------------
    # return

    return params



"""
This module provides a function to remove boundary tetras from tetrahedral meshes

"""

import os

import numpy as np
from lapy import TetMesh, io

# -----------------------------------------------------------------------------
# MAIN FUNCTION


def removeBoundaryMask(params):
    # -------------------------------------------------------------------------
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Removing boundary tetras from mesh")
    print()

    # -------------------------------------------------------------------------
    # evaluate input

    VTKFile = os.path.join(params.OUTDIR, params.HEMI + ".tetra.vtk")
    PSOLFile = os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".tetra.psol")

    # -------------------------------------------------------------------------
    # load data

    tetMesh = TetMesh.read_vtk(VTKFile)

    lbl = io.read_vfunc(PSOLFile)
    lbl = np.array(lbl)

    vTail = tetMesh.v[lbl == params.LUTDICT["bndtail"],]
    vHead = tetMesh.v[lbl == params.LUTDICT["bndhead"],]

    # -------------------------------------------------------------------------
    # cutting surfaces based on point-cloud PCA

    vTaile, vTailc = np.linalg.eig(np.cov(vTail, rowvar=False))
    vHeade, vHeadc = np.linalg.eig(np.cov(vHead, rowvar=False))

    # need to order EVs

    vTailc = vTailc[:, np.flip(np.argsort(vTaile), axis=0)]
    vHeadc = vHeadc[:, np.flip(np.argsort(vHeade), axis=0)]
    vTaile = np.flip(np.sort(vTaile), axis=0)
    vHeade = np.flip(np.sort(vHeade), axis=0)

    # -------------------------------------------------------------------------
    # determine on which side of the plane a given point is

    # head

    # support vectors (move a little bit inwards, direction may differ); first,
    # determine if the richtungsvektor is pointing towards the center:

    if np.linalg.norm((np.mean(tetMesh.v, axis=0) - np.mean(vHead, axis=0))) > np.linalg.norm(
        np.mean(tetMesh.v, axis=0) - (np.mean(vHead, axis=0) + vHeadc[:, 2])
    ):  # yes
        sHead = np.mean(vHead, axis=0) + 1.0 * vHeadc[:, 2]
        dirHead = 1
    else:
        sHead = np.mean(vHead, axis=0) - 1.0 * vHeadc[:, 2]
        dirHead = 0

    # compute unit normal vector to plane

    uHead = vHeadc[:, 2] / np.linalg.norm(vHeadc[:, 2])

    # compute normals from each point to plane (could be v or vHead)

    dHead = np.zeros(tetMesh.v.shape[0])

    for i in range(0, tetMesh.v.shape[0]):
        # vector from point to support vector
        n = tetMesh.v[i, :] - sHead

        # distance from point to plane along normal
        dHead[i] = np.dot(n, uHead)

    # tail

    # support vectors (move a little bit inwards, direction may differ); first,
    # determine if the richtungsvektor is pointing towards the center:

    if np.linalg.norm((np.mean(tetMesh.v, axis=0) - np.mean(vTail, axis=0))) > np.linalg.norm(
        np.mean(tetMesh.v, axis=0) - (np.mean(vTail, axis=0) + vTailc[:, 2])
    ):
        sTail = np.mean(vTail, axis=0) + 1.0 * vTailc[:, 2]  # yes
        dirTail = 1
    else:
        sTail = np.mean(vTail, axis=0) - 1.0 * vTailc[:, 2]  # no
        dirTail = 0

    # compute unit normal vector to plane

    uTail = vTailc[:, 2] / np.linalg.norm(vTailc[:, 2])

    # compute normals from each point to plane (could be v or vTail)

    dTail = np.zeros(tetMesh.v.shape[0])

    for i in range(0, tetMesh.v.shape[0]):
        # vector from point to support vector
        n = tetMesh.v[i, :] - sTail

        # distance from point to plane along normal
        dTail[i] = np.dot(n, uTail)

    # -------------------------------------------------------------------------
    # remove triangles that contain 2 or 3 or 4 points

    if dirHead & dirTail:
        tcut = tetMesh.t[
            np.sum(np.reshape(np.in1d(tetMesh.t, np.where((dTail <= 0) | (dHead <= 0))), tetMesh.t.shape), axis=1) < 2,
            :,
        ]
        vcutIdxTail = np.where(dTail <= 0)[0]
        vcutIdxHead = np.where(dHead <= 0)[0]
    elif (not (dirHead)) & dirTail:
        tcut = tetMesh.t[
            np.sum(np.reshape(np.in1d(tetMesh.t, np.where((dTail <= 0) | (dHead > 0))), tetMesh.t.shape), axis=1) < 2, :
        ]
        vcutIdxTail = np.where(dTail <= 0)[0]
        vcutIdxHead = np.where(dHead > 0)[0]
    elif dirHead & (not (dirTail)):
        tcut = tetMesh.t[
            np.sum(np.reshape(np.in1d(tetMesh.t, np.where((dTail > 0) | (dHead <= 0))), tetMesh.t.shape), axis=1) < 2, :
        ]
        vcutIdxTail = np.where(dTail > 0)[0]
        vcutIdxHead = np.where(dHead <= 0)[0]
    elif (not (dirHead)) & (not (dirTail)):
        tcut = tetMesh.t[
            np.sum(np.reshape(np.in1d(tetMesh.t, np.where((dTail > 0) | (dHead > 0))), tetMesh.t.shape), axis=1) < 2, :
        ]
        vcutIdxTail = np.where(dTail > 0)[0]
        vcutIdxHead = np.where(dHead > 0)[0]

    # -------------------------------------------------------------------------
    # write PSOL

    vIdx = np.zeros(np.shape(tetMesh.v)[0])
    vIdx[vcutIdxHead] = params.LUTDICT["bndtail"]
    vIdx[vcutIdxTail] = params.LUTDICT["bndhead"]

    io.write_vfunc(os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".tetra-remove_bnd.psol"), vIdx)

    # for visualization

    tetMeshBnd = tetMesh.boundary_tria()

    vRmBndKeep, vBndRmDel = tetMeshBnd.rm_free_vertices_()

    tetMeshBnd.orient_()

    tetMeshBnd.write_vtk(filename=os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".rm.bnd.tetra.vtk"))

    io.write_vfunc(os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".rm.bnd.tetra.psol"), vIdx)

    # -------------------------------------------------------------------------
    # return

    return params



"""
This module provides a function to check surfaces

"""

import logging
import os

from lapy import TriaMesh

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def checkSurface(params, stage=None):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Check surfaces")
    print()

    if params.internal.CHECKSURFACE is True and stage == "check_surface":
        triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk"))

        euler = triaMesh.euler()

        LOGGER.info("Euler number for " + os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk") + " is " + str(euler))

        if euler != 2:
            LOGGER.info("Surface contains holes. Please edit the corresponding hippocampal segmentation and re-run.")
            continue_program = False

        else:
            continue_program = True

    elif params.internal.CHECKBOUNDARIES is True and stage == "check_boundaries":
        triaMesh = TriaMesh.read_vtk(
            os.path.join(
                os.path.join(params.OUTDIR, "tetra-cut"),
                params.HEMI + ".rm.open.bnd.cut.vtk",
            )
        )

        bnd_loops = triaMesh.boundary_loops()

        LOGGER.info(
            "There are "
            + str(len(bnd_loops))
            + " boundary loops for "
            + os.path.join(
                os.path.join(params.OUTDIR, "tetra-cut"),
                params.HEMI + ".rm.open.bnd.cut.vtk",
            )
        )

        if len(bnd_loops) != 2:
            LOGGER.info(
                "Surface contains does not contain 2 boundary loops. Please retry with different cutting parameters."
            )
            continue_program = False

        else:
            continue_program = True

    else:
        continue_program = True

    #

    params.internal.continue_program = continue_program

    #

    return params



"""
This module provides a function to create a label files for a tetrahedral mesh

"""

import os
import subprocess

import nibabel as nb
import numpy as np
from lapy import TetMesh, TriaMesh, io

# -----------------------------------------------------------------------------
# AUXILIARY FUNCTIONS

# createBoundaryMask()


def _createBoundaryMask(hbtFile, mskFile, outFile, label, bndlabel):
    """
    createBoundaryMask(hbtFile, mskFile)

    """

    # load segmentations
    hbt = nb.load(hbtFile)
    msk = nb.load(mskFile)

    # define new data structures
    bnd = msk
    bnd_data = bnd.get_fdata(caching="fill")
    hbt_data_flat = hbt.get_fdata().flatten()

    # find nonzero elements of msk
    mskx, msky, mskz = np.where(msk.get_fdata())

    # cycle through each nonzero element of msk
    for i in range(0, len(mskx)):
        # create indices xyz of local neighborhood (one more +1 due to python indexing)
        mx, my, mz = np.meshgrid(
            range(mskx[i] - 1, mskx[i] + 1 + 1),
            range(msky[i] - 1, msky[i] + 1 + 1),
            range(mskz[i] - 1, mskz[i] + 1 + 1),
        )

        # convert xyz to i
        # sel = range(0,27) # point conn
        # sel = range(0,len(mx.flatten())) # use this if increasing the neighborhood size above, e.g. to 3 or 6
        sel = np.array((5, 11, 13, 14, 15, 17, 23)) - 1  # face conn
        m = np.ravel_multi_index([mx.flatten()[sel], my.flatten()[sel], mz.flatten()[sel]], msk.shape)

        # get values from hbt and set bnd to 2260 if it contains any tail voxel
        # (226) and to 2320 if it contains any head voxel (232)

        for j in range(0, len(label)):
            if any(np.isin(hbt_data_flat[m], label[j])):
                bnd_data[mskx[i], msky[i], mskz[i]] = bndlabel[j]

    # write boundary mask; since we have used 'caching='fill'' while loading
    # the data, the contents of bnd will be automatically updated via bnd_data
    out = nb.MGHImage(bnd_data, affine=bnd.affine, header=bnd.header)
    nb.freesurfer.save(out, outFile)


# ------------------------------------------------------------------------------
# MAIN FUNCTION

# createTetraLabels


def createTetraLabels(params):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Creating label files for tetrahedral meshes")
    print()

    # determine files

    MSKFILE = os.path.join(params.OUTDIR, params.HEMI + ".labels.mgz")

    if params.internal.AUTOMASK_HEAD is True or params.internal.AUTOMASK_TAIL is True:
        HBTFILE = os.path.join(params.OUTDIR, "labels", params.HEMI + ".automask_image.mgz")
    else:
        HBTFILE = os.path.join(params.OUTDIR, params.HEMI + ".image.mgz")

    # create new masks that includes boundary labels to head and tail, and to CA4

    _createBoundaryMask(
        hbtFile=HBTFILE,
        mskFile=MSKFILE,
        outFile=os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + "." + "labels-bnd.mgz"),
        label=[params.LUTDICT["tail"], params.LUTDICT["head"]],
        bndlabel=[params.LUTDICT["bndtail"], params.LUTDICT["bndhead"]],
    )
    _createBoundaryMask(
        hbtFile=HBTFILE,
        mskFile=MSKFILE,
        outFile=os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + "." + "labels-ca4.mgz"),
        label=[params.LUTDICT["ca4"]],
        bndlabel=[params.LUTDICT["bndca4"]],
    )

    # merge bnd and ca4 masks (add ca4 to bnd)

    cmd = (
        os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mris_calc")
        + " "
        + "--output "
        + os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + "." + "labels-bnd.mgz")
        + " "
        + os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + "." + "labels-bnd.mgz")
        + " "
        + "lowerlimit "
        + os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + "." + "labels-ca4.mgz")
    )

    print(cmd)

    subprocess.run(cmd.split(), capture_output=True)

    # load bnd image

    bndImage = nb.freesurfer.load(os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".labels-bnd.mgz"))

    # load tet mesh

    tetMesh = TetMesh.read_vtk(os.path.join(params.OUTDIR, params.HEMI + ".tetra.vtk"))

    # convert vertex indices to voxel indices

    vtxVxIdx = np.round(
        np.matmul(
            np.linalg.inv(bndImage.header.get_vox2ras_tkr()),
            np.concatenate((tetMesh.v, np.ones((tetMesh.v.shape[0], 1))), axis=1).T,
        ).T
    ).astype(int)[:, 0:3]

    # lookup labels of converted vertex indices in bnd image

    vtxLabels = bndImage.get_fdata()[vtxVxIdx[:, 0], vtxVxIdx[:, 1], vtxVxIdx[:, 2]]

    # write out psol file

    with open(os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".tetra.psol"), "w") as f:
        print("Solution:", file=f)
        [print(str(int(x)), file=f) for x in vtxLabels]

    # write out visualization files

    tetMeshBnd = tetMesh.boundary_tria()
    vRmBndKeep, vBndRmDel = tetMeshBnd.rm_free_vertices_()
    tetMeshBnd.orient_()
    TriaMesh.write_vtk(tetMeshBnd, os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".rm.bnd.tetra.vtk"))
    io.write_vfunc(
        os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".rm.bnd.tetra.psol"),
        np.array([float(x) for x in vtxLabels])[vRmBndKeep],
    )
    nb.freesurfer.save(
        nb.freesurfer.MGHImage(
            dataobj=np.array([float(x) for x in vtxLabels])[vRmBndKeep].astype("float32"), affine=None
        ),
        filename=os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".rm.bnd.label.mgh"),
    )

    # return

    return params



"""
This module provides the main functionality of the hippocampal shape and
thickness analysis package.

"""

import argparse
import logging
import os
import shutil
import sys
import time

from .cfg.atlases import get_atlases
from .cfg.config import get_defaults
from .cfg.logging import setup_logging
from .cfg.version import get_version
from .computeCubeParam import computeCubeParam
from .computeThickness import computeThickness
from .createSurface import extractSurface, remeshSurface, smoothSurface
from .createTetraLabels import createTetraLabels
from .createTetraMesh import createTetraMesh
from .cutTetra import cutTetra
from .doc.documentation import get_help_text
from .processImage import convertFormat, copy_image_to_main, cropImage, upsampleImage
from .processLabels import autoMask, copy_labels_to_main, createLabels, mergeMolecularLayer
from .processMask import binarizeMask, closeMask, copy_mask_to_main, gaussFilter, longFilter
from .removeBoundaryMask import removeBoundaryMask
from .utils.check_surface import checkSurface
from .utils.create_supplementary_files import createSupplementaryFiles
from .utils.map_values import mapValues
from .utils.qc_plots import qcPlots

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS

# ------------------------------------------------------------------------------
# get_help()


def get_help(print_help=True, return_help=False):
    """
    a function to return a help message

    """

    HELPTEXT = get_help_text()

    if print_help:
        print(HELPTEXT)

    if return_help:
        return HELPTEXT


# ------------------------------------------------------------------------------
# check environment and packages


def _check_environment_and_packages():
    # check environment variables
    if os.environ.get("FREESURFER_HOME") is None:
        raise RuntimeError("Need to set the FreeSurfer_HOME environment variable")

    # check python version
    if sys.version_info <= (3, 5):
        raise RuntimeError("Python version must be 3.5 or greater")

    # check for gmsh
    if shutil.which("gmsh") is None:
        raise RuntimeError("Could not find a 'gmsh' executable")


# ------------------------------------------------------------------------------
# check directories


def _create_directories(args):
    # note that the main output directory has been created during
    # _start_logging() already

    # define list of directories

    list_of_directories = [
        os.path.join(args.outputdir, "image"),
        os.path.join(args.outputdir, "labels"),
        os.path.join(args.outputdir, "mask"),
        os.path.join(args.outputdir, "surface"),
        os.path.join(args.outputdir, "tetra-mesh"),
        os.path.join(args.outputdir, "tetra-labels"),
        os.path.join(args.outputdir, "tetra-cut"),
        os.path.join(args.outputdir, "tetra-cube"),
        os.path.join(args.outputdir, "thickness"),
        os.path.join(args.outputdir, "qc"),
    ]

    # loop over list of directories

    for directory in list_of_directories:
        if not os.path.isdir(directory):
            try:
                LOGGER.info("Creating output directory " + directory)
                os.mkdir(directory)
            except OSError as e:
                LOGGER.error("Cannot create output directory " + args.outputdir)
                LOGGER.error("Reason: " + str(e))
                raise


# ------------------------------------------------------------------------------
# parse_arguments


def _parse_arguments():
    """
    an internal function to parse input arguments

    """

    # setup parser
    parser = argparse.ArgumentParser(
        description="This program conducts a thickness analysis of the hippocampus, based on a FreeSurfer, ASHS, or custom hippocampal subfield segmentation.",
        add_help=False,
    )

    # required arguments (set to required=False, because not required when using --more-help)
    required = parser.add_argument_group("Required arguments")

    required.add_argument(
        "--filename",
        dest="filename",
        help="Filename of a segmentation file.",
        default=None,
        metavar="<filename>",
        required=False,
    )
    required.add_argument(
        "--hemi", dest="hemi", help="Hemisphere. Either 'lh' or 'rh'.", default=None, metavar="<lh|rh>", required=False
    )
    required.add_argument(
        "--lut",
        dest="lut",
        help="Look-up table: a text file with numeric and verbal segmentation labels. 'freesurfer', 'ashs-penn_abc_3t_t2', and 'ashs-umcutrecht_7t' are keywords for built-in tables.",
        default=None,
        metavar="<freesurfer|ashs-penn_abc_3t_t2|ashs-umcutrecht_7t|filename>",
        required=False,
    )
    required.add_argument(
        "--outputdir",
        dest="outputdir",
        help="Directory where the results will be written.",
        default=None,
        metavar="<directory>",
        required=False,
    )

    # optional arguments
    optional = parser.add_argument_group("Optional arguments")

    optional.add_argument(
        "--no-cleanup",
        dest="no_cleanup",
        help="Do not remove files that may be useful for diagnostic or debugging purposes, but are not necessary otherwise.",
        default=get_defaults("no_cleanup"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--no-crop",
        dest="no_crop",
        help="Do not crop image.",
        default=get_defaults("no_crop"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--upsample",
        dest="upsample",
        help="Upsample to the smallest voxel edge length.",
        default=get_defaults("upsample"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--upsample-size",
        dest="upsample_size",
        help="Upsampling factors. Should be between 0 and 1. If all zeros, upsample to the smallest voxel edge length. Default: 0 0 0",
        default=get_defaults("upsample_size"),
        metavar="<float>",
        nargs=3,
        required=False,
        type=float,
    )
    optional.add_argument(
        "--no-merge-molecular-layer",
        dest="no_merge_molecular_layer",
        help="Do not merge molecular layer (only applicable for FreeSurfer segmentations).",
        default=get_defaults("no_merge_molecular_layer"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--automask-head",
        dest="automask_head",
        help="Automated boundary detection for hippocampal head.",
        default=get_defaults("automask_head"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--automask-tail",
        dest="automask_tail",
        help="Automated boundary detection for hippocampal tail.",
        default=get_defaults("automask_tail"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--automask-head-margin",
        dest="automask_head_margin",
        help="Margin for automated boundary detection for hippocampal head. Default: 0",
        default=get_defaults("automask_head_margin"),
        metavar="<int>",
        required=False,
        type=int,
    )
    optional.add_argument(
        "--automask-tail-margin",
        dest="automask_tail_margin",
        help="Margin for automated boundary detection for hippocampal tail. Default: 0",
        default=get_defaults("automask_tail_margin"),
        metavar="<int>",
        required=False,
        type=int,
    )
    optional.add_argument(
        "--no-gauss-filter",
        dest="no_gauss_filter",
        help="Do not apply gaussian filter.",
        default=get_defaults("no_gauss_filter"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--gauss-filter-size",
        dest="gauss_filter_size",
        help="Filter width and threshold for gaussian filtering. Default: 1 50.",
        default=get_defaults("gauss_filter_size"),
        metavar="<float>",
        nargs=2,
        required=False,
        type=float,
    )
    optional.add_argument(
        "--long-filter",
        dest="long_filter",
        help="Apply filter along longitudinal axis, i.e. attempt to create smooth transitions between slices.",
        default=get_defaults("long_filter"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--long-filter-size",
        dest="long_filter_size",
        help="Size of longitudinal filter. Default: 5",
        default=get_defaults("long_filter_size"),
        metavar="<int>",
        required=False,
        type=int,
    )
    optional.add_argument(
        "--no-close-mask",
        dest="no_close_mask",
        help="Do not apply closing operation to mask, i.e. do not attempt to close small holes.",
        default=get_defaults("no_close_mask"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--mca",
        dest="mca",
        help="Type of marching-cube algorithm. Either 'mri_mc' or 'skimage'. Default: 'skimage'",
        default=get_defaults("mca"),
        metavar="<mri_mc|skimage>",
        required=False,
    )
    optional.add_argument(
        "--remesh",
        dest="remesh",
        help="Apply remeshing operation to surface, i.e. create a regular, evenly spaced surface grid.",
        default=get_defaults("remesh"),
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--smooth",
        dest="smooth",
        help="Mesh smoothing iterations. Default: 5",
        default=get_defaults("smooth"),
        metavar="<int>",
        required=False,
        type=int,
    )
    optional.add_argument(
        "--cut-range",
        dest="cut_range",
        help="Range for tetrahedral boundary cutting. Default: -0.975, 0.975",
        default=get_defaults("cut_range"),
        metavar="<float>",
        nargs=2,
        required=False,
        type=float,
    )
    optional.add_argument(
        "--aniso-alpha",
        dest="aniso_alpha",
        help="Anisotropy parameter(s). Can be one or two or two numbers. Default: 40",
        default=get_defaults("aniso_alpha"),
        metavar="<float>",
        nargs="+",
        required=False,
        type=float,
    )
    optional.add_argument(
        "--aniso-smooth",
        dest="aniso_smooth",
        help="Anisotropy smoothing iterations. Default: 3",
        default=get_defaults("aniso_smooth"),
        metavar="<int>",
        required=False,
        type=int,
    )
    optional.add_argument(
        "--thickness-grid",
        dest="thickness_grid",
        help="Extent and resolution of the grid used for thickness computation; three lists of three numbers: negative extent of x axis, positive extent of x axis, resolution on x axis. Repeat for the y and z axes. Default: -0.9 0.9 41 -0.975 0.975 21 -0.9 0.9 11",
        default=get_defaults("thickness_grid"),
        metavar="<float>",
        nargs=9,
        required=False,
        type=float,
    )

    # expert options
    expert = parser.add_argument_group("Expert options")

    expert.add_argument(
        "--mcc",
        dest="mcc",
        help=argparse.SUPPRESS,
        default=get_defaults("mcc"),
        metavar="<int>",
        required=False,
        type=int,
    )  # help="Marching-cube connectivity. Only used for \'mri_mc\' algorithm. Default: 1",
    expert.add_argument(
        "--remesh-size",
        dest="remesh_size",
        help=argparse.SUPPRESS,
        default=get_defaults("remesh_size"),
        metavar="<int>",
        required=False,
        type=int,
    )  # help="Target number of vertices for remeshing operation. If zero, keep original number of vertices. Default: 0",
    expert.add_argument(
        "--no-check-surface",
        dest="no_check_surface",
        help=argparse.SUPPRESS,
        default=get_defaults("no_check_surface"),
        action="store_true",
        required=False,
    )  # help="Do not check surface and proceed if there are holes.",
    expert.add_argument(
        "--no-check-boundaries",
        dest="no_check_boundaries",
        help=argparse.SUPPRESS,
        default=get_defaults("no_check_boundaries"),
        action="store_true",
        required=False,
    )  # help="Do not check boundaries and proceed if there are less / more than two continuous boundary loops.",
    expert.add_argument(
        "--no-qc",
        dest="no_qc",
        help=argparse.SUPPRESS,
        default=get_defaults("no_qc"),
        action="store_true",
        required=False,
    )  # help="Do not perform QC",
    expert.add_argument(
        "--allow-ragged-surfaces",
        dest="allow_ragged_surfaces",
        help=argparse.SUPPRESS,
        default=get_defaults("allow_ragged_surfaces"),
        action="store_true",
        required=False,
    )  # help="Allow ragged mid-surfaces.",
    expert.add_argument(
        "--allow-ragged-trias",
        dest="allow_ragged_trias",
        help=argparse.SUPPRESS,
        default=get_defaults("allow_ragged_trias"),
        action="store_true",
        required=False,
    )  # help="Allow triangles for ragged mid-surfaces.",

    # deprecated options - to be removed in the future
    deprecated = parser.add_argument_group("Deprecated options")

    deprecated.add_argument(
        "--no-orient",
        dest="no_orient",
        help=argparse.SUPPRESS,
        default=get_defaults("no_orient"),
        action="store_true",
        required=False,
    )  # help="Do not check surface.",

    # define help
    help = parser.add_argument_group("Getting help")

    help.add_argument("--help", help="Display this help message and exit", action="help")
    help.add_argument(
        "--more-help",
        dest="more_help",
        help="Display extensive help message and exit",
        default=False,
        action="store_true",
        required=False,
    )
    help.add_argument(
        "--version", help="Display version number and exit", action="version", version="%(prog)s " + get_version()
    )

    # check if there are any inputs; if not, print help
    if len(sys.argv) == 1:
        args = parser.parse_args(["--help"])
    else:
        args = parser.parse_args()

    #
    if args.more_help is False:
        # check for required arguments, print message and exit
        if args.filename is None:
            print("ERROR: the --filename argument is required, exiting. Use --help to see details.")
            print()
            sys.exit(1)

        if args.hemi is None:
            print("ERROR: the --hemi argument is required, exiting. Use --help to see details.")
            print()
            sys.exit(1)

        if args.lut is None:
            print("ERROR: the --lut argument is required, exiting. Use --help to see details.")
            print()
            sys.exit(1)

        if args.outputdir is None:
            print("ERROR: the --outputdir argument is required, exiting. Use --help to see details.")
            print()
            sys.exit(1)

    #
    return args


# ------------------------------------------------------------------------------
# evaluate arguments


def _evaluate_args(args):
    """
    an internal function to set and return internal parameters

    """

    # message

    LOGGER.info("Evaluating arguments ...")

    # initialize settings

    class settings:
        pass

    # clean up

    settings.CLEANUP = not (args.no_cleanup)

    # processImage

    settings.CROP = not (args.no_crop)
    settings.UPSAMPLE = args.upsample
    settings.UPSAMPLE_SIZE = args.upsample_size

    # processLabels

    ## mergeMolecularLayer

    if args.lut == "freesurfer" and args.no_merge_molecular_layer is False:
        settings.MERGE_MOLECULAR_LAYER = True
    elif args.lut == "freesurfer" and args.no_merge_molecular_layer is True:
        settings.MERGE_MOLECULAR_LAYER = False
    else:
        settings.MERGE_MOLECULAR_LAYER = False

    ## automask

    settings.AUTOMASK_HEAD = args.automask_head
    settings.AUTOMASK_TAIL = args.automask_tail
    settings.AUTOMASK_HEAD_MARGIN = args.automask_head_margin
    settings.AUTOMASK_TAIL_MARGIN = args.automask_tail_margin

    # processMask

    settings.GAUSSFILTER = not (args.no_gauss_filter)
    settings.GAUSSFILTER_SIZE = args.gauss_filter_size
    settings.LONGFILTER = args.long_filter
    settings.LONGFILTER_SIZE = args.long_filter_size
    settings.CLOSEMASK = not (args.no_close_mask)

    # createSurface

    ## marching cube algorithm

    settings.MCA = args.mca

    ## marching cube connectivity; 1=6+, 2=18, 3=6, 4=26

    settings.MCC = args.mcc

    ## remeshing

    settings.REMESH = args.remesh
    settings.REMESH_SIZE = args.remesh_size

    ## smoothing iterations

    settings.SMO = args.smooth

    # qc params

    settings.no_qc = args.no_qc

    # checkSurface

    settings.CHECKSURFACE = not (args.no_check_surface)
    settings.CHECKBOUNDARIES = not (args.no_check_boundaries)

    # cutTetra

    settings.cut_range = args.cut_range

    # computeCubeParam

    settings.aniso_alpha = args.aniso_alpha
    settings.aniso_smooth = args.aniso_smooth

    # computeThickness

    settings.THXn = args.thickness_grid[0]
    settings.THXp = args.thickness_grid[1]
    settings.THXk = int(args.thickness_grid[2])
    settings.THYn = args.thickness_grid[3]
    settings.THYp = args.thickness_grid[4]
    settings.THYk = int(args.thickness_grid[5])
    settings.THZn = args.thickness_grid[6]
    settings.THZp = args.thickness_grid[7]
    settings.THZk = int(args.thickness_grid[8])
    settings.allow_ragged_surfaces = args.allow_ragged_surfaces
    settings.allow_ragged_trias = args.allow_ragged_trias
    settings.no_orient = args.no_orient

    # mapValues

    settings.mapValuesIntegrate = get_defaults("map_values_integrate")
    settings.mapValuesSelect = get_defaults("map_values_select")
    settings.mapValuesInterp = get_defaults("map_values_interp")
    settings.mapValuesWritePSOL = get_defaults("map_values_write_psol")
    settings.mapValuesWriteMGH = get_defaults("map_values_write_mgh")
    settings.mapValuesWriteANNOT = get_defaults("map_values_write_annot")

    # create the LUT

    LUTDICT, HSFLIST = get_atlases(args.lut)

    # assemble and return params

    class params:
        pass

    params.FILENAME = os.path.abspath(args.filename)
    params.HEMI = args.hemi
    params.LUT = args.lut
    params.OUTDIR = args.outputdir

    params.LUTDICT = LUTDICT
    params.HSFLIST = HSFLIST

    params.internal = settings

    LOGGER.info("Using " + params.FILENAME + " as input file")
    LOGGER.info("Using " + params.OUTDIR + " as output directory")

    return params


# ------------------------------------------------------------------------------
# check params


def _check_params(params):
    """
    an internal function to check input arguments

    """

    # check for subfield segmentation file

    if os.path.isfile(params.FILENAME):
        LOGGER.info("Found " + params.FILENAME)
    else:
        raise RuntimeError("Could not find " + params.FILENAME)

    # check hemisphere

    if params.HEMI != "lh" and params.HEMI != "rh":
        raise RuntimeError("Hemisphere must be either lh or rh, but not " + params.HEMI + ", exiting.")

    # check MC algorithm

    if params.internal.MCA != "mri_mc" and params.internal.MCA != "mri_tessellate" and params.internal.MCA != "skimage":
        raise RuntimeError("Could not recognise marching cube algorithm " + params.internal.MCA + ", exiting.")

    # check upsampling

    if any(x < 0 for x in params.internal.UPSAMPLE_SIZE) or any(x > 1 for x in params.internal.UPSAMPLE_SIZE):
        raise RuntimeError("All upsampling parameters should be between 0 and 1.")

    # check ML

    if params.internal.MERGE_MOLECULAR_LAYER is True and params.LUT != "freesurfer":
        raise RuntimeError("Cannot use no-merge-molecular-layer with " + params.LUT + ".")

    if params.internal.MERGE_MOLECULAR_LAYER is True and ("ml" in params.LUTDICT.keys()) is False:
        LOGGER.info("Could not find molecular layer in lookup table.")
        raise AssertionError()

    # check longfilter size

    if params.internal.LONGFILTER_SIZE < 2 or (params.internal.LONGFILTER_SIZE % 2) == 0:
        raise RuntimeError("Longfilter size must be an uneven integer greater than 2.")

    # check aniso alpha

    if len(params.internal.aniso_alpha) > 2:
        raise RuntimeError("Length of aniso-alpha must be 1 or 2.")

    # check LUT

    if (
        params.LUT != "freesurfer"
        and params.LUT != "ashs-penn_abc_3t_t2"
        and params.LUT != "ashs-umcutrecht_7t"
        and not os.path.isfile(params.LUT)
    ):
        raise RuntimeError(
            "Look-up table can only be 'freesurfer', 'ashs-penn_abc_3t_t2', 'ashs-umcutrecht_7t', or an existing file, but not "
            + params.LUT
        )

    # return

    return params


# ------------------------------------------------------------------------------
# run analysis


def _run_analysis(params):
    """ """

    # process image (1)

    LOGGER.info("Starting convertFormat() ...")
    params = convertFormat(params)

    LOGGER.info("Starting cropImage() ...")
    params = cropImage(params)

    LOGGER.info("Starting upsampleImage() ...")
    params = upsampleImage(params)

    LOGGER.info("Starting copy_image_to_main() ...")
    params = copy_image_to_main(params)

    # process labels (2)

    LOGGER.info("Starting autoMask() ...")
    params = autoMask(params)

    LOGGER.info("Starting createLabels() ...")
    params = createLabels(params)

    LOGGER.info("Starting mergeMolecularLayer() ...")
    params = mergeMolecularLayer(params)

    LOGGER.info("Starting copy_labels_to_main() ...")
    params = copy_labels_to_main(params)

    # process mask (3)

    LOGGER.info("Starting binarizeMask() ...")
    params = binarizeMask(params)

    LOGGER.info("Starting gaussFilter() ...")
    params = gaussFilter(params)

    LOGGER.info("Starting longFilter() ...")
    params = longFilter(params)

    LOGGER.info("Starting closeMask() ...")
    params = closeMask(params)

    LOGGER.info("Starting copy_mask_to_main() ...")
    params = copy_mask_to_main(params)

    # create surface (4)

    LOGGER.info("Starting extractSurface() ...")
    params = extractSurface(params)

    LOGGER.info("Starting remeshSurface() ...")
    params = remeshSurface(params)

    LOGGER.info("Starting smoothSurface() ...")
    params = smoothSurface(params)

    LOGGER.info("Starting qcPlots() ...")
    params = qcPlots(params, stage="mesh")

    LOGGER.info("Starting checkSurface() ...")
    params = checkSurface(params, stage="check_surface")

    if params.internal.continue_program is False:
        LOGGER.info("Hipsta finished WITH ERRORS.")
        raise AssertionError("Check surface failed (stage: surface)")

    # create tetra mesh for whole hippocampal body (5)

    LOGGER.info("Starting createTetraMesh() ...")
    params = createTetraMesh(params)

    # create label files for tetra mesh (6)

    LOGGER.info("Starting createTetraLabels() ...")
    params = createTetraLabels(params)

    # cut hippocampal body (7)

    LOGGER.info("Starting removeBoundaryMask() ...")
    params = removeBoundaryMask(params)

    LOGGER.info("Starting cutTetra() ...")
    params = cutTetra(params)

    LOGGER.info("Starting checkSurface() ...")
    params = checkSurface(params, stage="check_boundaries")

    if params.internal.continue_program is False:
        LOGGER.info("Hipsta finished WITH ERRORS.")
        raise AssertionError("Check surface failed (stage: boundaries)")

    # compute cube parametrization (8)

    LOGGER.info("Starting computeCubeParam() ...")
    params = computeCubeParam(params)

    LOGGER.info("Starting qcPlots() ...")
    params = qcPlots(params, stage="profile")

    # compute thickness (9)

    LOGGER.info("Starting computeThickness() ...")
    params = computeThickness(params)

    LOGGER.info("Starting qcPlots() ...")
    params = qcPlots(params, stage="hull")

    # map subfield mapValues (10)

    LOGGER.info("Starting mapValues() ...")
    params = mapValues(params)

    # create supplementary files (11)

    LOGGER.info("Starting createSupplementaryFiles() ...")
    params = createSupplementaryFiles(params)

    # all done

    LOGGER.info("Date: %s", time.strftime("%d/%m/%Y %H:%M:%S"))
    LOGGER.info("Hipsta finished without errors.")


# ------------------------------------------------------------------------------
# _run_hipsta


def _run_hipsta(args):
    """
    a function to run the hipsta submodules

    """

    #
    try:
        #
        setup_logging(args)

        # check environment and packages
        _check_environment_and_packages()

        # create directories
        _create_directories(args)

        # convert arguments to params
        params = _evaluate_args(args)

        # check params
        _check_params(params)

        # run analysis
        _run_analysis(params)

    except Exception as e:
        LOGGER.error("Error Information:")
        LOGGER.error("Type: %s", type(e))
        LOGGER.error("Value: %s", e.args)

        raise


# ------------------------------------------------------------------------------
# run_hipsta


def run_hipsta(filename, hemi, lut, outputdir, **kwargs):
    """
    Run the hippocampal shape and thickness analysis

    Parameters
    ----------

    filename :
        Filename of a segmentation file.
    hemi :
        Hemisphere. Either \'lh\' or \'rh\'.
    lut :
        Look-up table: a text file with numeric and verbal segmentation labels. \'freesurfer\', 'ashs-penn_abc_3t_t2' and \'ashs-umcutrecht_7t\' are keywords for built-in tables.
    outputdir
        Directory where the results will be written.

    no_cleanup : optional
        Keep files that may be useful for diagnostic or debugging purposes, but are not necessary otherwise. Default: False
    no_crop : optional
        Do not crop image. Default: False.
    upsample : optional
        Upsample to the smallest voxel edge length. Default: False
    upsample_size : optional
        Upsampling factors. Should be between 0 and 1. If all zeros, upsample to the smallest voxel edge length. Default: [0, 0, 0]
    no_merge_molecular_layer : optional
        Do not merge molecular layer (only applicable for FreeSurfer segmentations). Default: False
    automask_head : optional
        Automated boundary detection for hippocampal head. Default: False
    automask_tail : optional
        Automated boundary detection for hippocampal tail. Default: False
    automask_head_margin : optional
        Margin for automated boundary detection for hippocampal head. Default: 0
    automask_tail_margin : optional
        Margin for automated boundary detection for hippocampal tail. Default: 0
    no_gauss_filter : optional
        Do not apply gaussian filter. Default: False
    gauss_filter_size : optional
        Filter width and threshold for gaussian filtering. Default: [1, 50]
    long_filter : optional
        Apply filter along longitudinal axis, i.e. attempt to create smooth transitions between slices. Default: False
    long_filter_size : optional
        Size of longitudinal filter. Default: 5
    no_close_mask : optional
        Do not apply closing operation to mask, i.e. do not attempt to close small holes. Default: False
    mca : optional
        Type of marching-cube algorithm. Either \'mri_mc\' or \'skimage\'. Default: \'mri_mc\'
    remesh : optional
        Apply remeshing operation to surface, i.e. create a regular, evenly spaced surface grid. Default: False
    smooth : optional
        Mesh smoothing iterations. Default: 5
    cut_range : optional
        Range for tetrahedral boundary cutting. Default: [-0.975, 0.975]
    aniso_alpha : optional
        Anisotropy parameter(s). Can be one or two numbers. Default: [40]
    aniso_smooth : optional
        Anisotropy smoothing iterations parameter. Default: 3
    thickness_grid : optional
        Extent and resolution of the grid used for thickness computation; three lists of three numbers: negative extent of x axis, positive extent of x axis, resolution on x axis. Repeat for the y and z axes. Default: [ -0.9, 0.9, 41, -0.975, 0.975, 21, -0.9, 0.9, 11]

    mcc : optional
        Marching-cube connectivity. Only used for \'mri_mc\' algorithm. Default: 1
    remesh_size : optional
        Target number of vertices for remeshing operation. If zero, keep original number of vertices. Default: 0
    no_check_surface : optional
        Do not check surface and proceed if there are holes. Default: False
    no_check_boundaries : optional
        Do not check boundaries and proceed if there are less / more than two continuous boundary loops. Default: False
    no_qc : optional
        Do not perform QC. Default: False
    allow_ragged_surfaces
        Allow ragged mid-surfaces. Default: False
    allow_ragged_trias
        Allow triangles for ragged mid-surfaces. Default: False

    no_orient : optional
        Do not orient surfaces. Default: False

    Returns
    -------
    dict
        a dictionary of input arguments
    """

    # define class

    class Args:
        def __init__(self, dct=None):
            # get defaults
            self.no_cleanup = get_defaults("no_cleanup")
            self.no_crop = get_defaults("no_crop")
            self.upsample = get_defaults("upsample")
            self.upsample_size = get_defaults("upsample_size")
            self.no_merge_molecular_layer = get_defaults("no_merge_molecular_layer")
            self.automask_head = get_defaults("automask_head")
            self.automask_tail = get_defaults("automask_tail")
            self.automask_head_margin = get_defaults("automask_head_margin")
            self.automask_tail_margin = get_defaults("automask_tail_margin")
            self.no_gauss_filter = get_defaults("no_gauss_filter")
            self.gauss_filter_size = get_defaults("gauss_filter_size")
            self.long_filter = get_defaults("long_filter")
            self.long_filter_size = get_defaults("long_filter_size")
            self.no_close_mask = get_defaults("no_close_mask")
            self.mca = get_defaults("mca")
            self.remesh = get_defaults("remesh")
            self.smooth = get_defaults("smooth")
            self.cut_range = get_defaults("cut_range")
            self.aniso_alpha = get_defaults("aniso_alpha")
            self.aniso_smooth = get_defaults("aniso_smooth")
            self.thickness_grid = get_defaults("thickness_grid")
            # expert options
            self.mcc = get_defaults("mcc")
            self.remesh_size = get_defaults("remesh_size")
            self.no_check_surface = get_defaults("no_check_surface")
            self.no_check_boundaries = get_defaults("no_check_boundaries")
            self.no_qc = get_defaults("no_qc")
            self.allow_ragged_surfaces = get_defaults("allow_ragged_surfaces")
            self.allow_ragged_trias = get_defaults("allow_ragged_trias")
            # deprecated options
            self.no_orient = get_defaults("no_orient")

            # parse kwargs
            if dct is not None:
                for key, value in dct.items():
                    if key in vars(self).keys():
                        setattr(self, key, value)
                    else:
                        raise AssertionError("Key " + key + " not included in list of default keys")

    #

    my_args = Args(kwargs)

    my_args.filename = filename
    my_args.hemi = hemi
    my_args.lut = lut
    my_args.outputdir = outputdir

    #

    _run_hipsta(my_args)

    #

    return vars(my_args)



"""
This module provides functions for processing label images

"""

import logging
import os
import shutil
import subprocess

import nibabel as nb
import numpy as np

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def autoMask(params):
    """ """

    if params.internal.AUTOMASK_HEAD is True or params.internal.AUTOMASK_TAIL is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Creating auto-mask")
        print()

        # get image
        img = nb.freesurfer.load(params.FILENAME)

        # get image data
        dat = img.get_fdata()

        # get orientation
        imgDims = nb.aff2axcodes(img.affine)

        # get dimension (reference is RAS)
        imgDimsAP = np.where(np.isin(imgDims, ["A", "P"]))[0][0]

        # get direction
        if imgDims[imgDimsAP] == "A":
            imgDimsAPDir = 1
        elif imgDims[imgDimsAP] == "P":
            imgDimsAPDir = -1

        # get LUT indices for head, tail and CA2/CA3
        labelHead = params.LUTDICT["head"]
        labelTail = params.LUTDICT["tail"]
        labelSbc = params.LUTDICT["subiculum"]
        labelCA1 = params.LUTDICT["ca1"]
        labelCA2 = params.LUTDICT["ca2"]
        labelCA3 = params.LUTDICT["ca3"]
        if "parahippocampal" in params.LUTDICT.keys():
            labelPHC = params.LUTDICT["parahippocampal"]
        else:
            labelPHC = None
        if "entorhinal" in params.LUTDICT.keys():
            labelENT = params.LUTDICT["entorhinal"]
        else:
            labelENT = None
        if "ba35" in params.LUTDICT.keys():
            labelBA35 = params.LUTDICT["ba35"]
        else:
            labelBA35 = None
        if "ba36" in params.LUTDICT.keys():
            labelBA36 = params.LUTDICT["ba36"]
        else:
            labelBA36 = None

        # head
        if params.internal.AUTOMASK_HEAD is True:
            #
            LOGGER.info("Using auto-mask for head")
            # remove any potentially existing head labels
            dat[dat == labelHead] = 0
            #
            if labelCA2 is not None and labelCA3 is not None:
                # find instance of CA2/CA3
                idxHead = np.argwhere(np.logical_or(dat == labelCA2, dat == labelCA3))
            else:
                raise RuntimeError("Insufficient label information, exiting.")
            # get min/max row/col/slice
            if imgDimsAPDir == -1:
                cutFrom = np.min(idxHead[:, imgDimsAP]) + params.internal.AUTOMASK_HEAD_MARGIN
            elif imgDimsAPDir == 1:
                cutFrom = np.max(idxHead[:, imgDimsAP]) - params.internal.AUTOMASK_HEAD_MARGIN
            # set to zero
            if imgDimsAP == 0 and imgDimsAPDir == -1:
                dat[:cutFrom, :, :] = 0
            elif imgDimsAP == 0 and imgDimsAPDir == 1:
                dat[cutFrom + 1 :, :, :] = 0
            elif imgDimsAP == 1 and imgDimsAPDir == -1:
                dat[:, :cutFrom, :] = 0
            elif imgDimsAP == 1 and imgDimsAPDir == 1:
                dat[:, cutFrom + 1 :, :] = 0
            elif imgDimsAP == 2 and imgDimsAPDir == -1:
                dat[:, :, :cutFrom] = 0
            elif imgDimsAP == 2 and imgDimsAPDir == 1:
                dat[:, :, cutFrom + 1 :] = 0
            # set to label
            if imgDimsAP == 0:
                dat[cutFrom, np.nonzero(dat[cutFrom, :, :])[0], np.nonzero(dat[cutFrom, :, :])[1]] = labelHead
            elif imgDimsAP == 1:
                dat[np.nonzero(dat[:, cutFrom, :])[0], cutFrom, np.nonzero(dat[:, cutFrom, :])[1]] = labelHead
            elif imgDimsAP == 2:
                dat[np.nonzero(dat[:, :, cutFrom])[0], np.nonzero(dat[:, :, cutFrom])[1], cutFrom] = labelHead

        # tail
        if params.internal.AUTOMASK_TAIL is True:
            #
            LOGGER.info("Using auto-mask for tail")
            # remove any potentially existing tail labels
            dat[dat == labelTail] = 0
            #
            if labelSbc is not None and labelCA1 is not None and labelCA2 is not None and labelCA3 is not None:
                ## find instance of Sbc/CA1/CA2/CA3
                # idxTail = np.intersect1d(
                #    np.argwhere(dat == labelSbc)[:, imgDimsAP],
                #    np.intersect1d(
                #        np.argwhere(dat == labelCA1)[:, imgDimsAP],
                #        np.intersect1d(
                #            np.argwhere(dat == labelCA2)[:, imgDimsAP], np.argwhere(dat == labelCA3)[:, imgDimsAP]
                #        ),
                #    ),
                # )
                # find instance of Sbc/CA1
                idxTail = np.intersect1d(
                    np.argwhere(dat == labelSbc)[:, imgDimsAP], np.argwhere(dat == labelCA1)[:, imgDimsAP]
                )
            else:
                raise RuntimeError("Insufficient label information, exiting.")
            # get min/max row/col/slice
            if imgDimsAPDir == -1:
                cutFrom = np.max(idxTail) - params.internal.AUTOMASK_TAIL_MARGIN
            elif imgDimsAPDir == 1:
                cutFrom = np.min(idxTail) + params.internal.AUTOMASK_TAIL_MARGIN
            # set to zero
            if imgDimsAP == 0 and imgDimsAPDir == -1:
                dat[cutFrom + 1 :, :, :] = 0
            elif imgDimsAP == 0 and imgDimsAPDir == 1:
                dat[:cutFrom, :, :] = 0
            elif imgDimsAP == 1 and imgDimsAPDir == -1:
                dat[:, cutFrom + 1 :, :] = 0
            elif imgDimsAP == 1 and imgDimsAPDir == 1:
                dat[:, :cutFrom:, :] = 0
            elif imgDimsAP == 2 and imgDimsAPDir == -1:
                dat[:, :, cutFrom + 1 :] = 0
            elif imgDimsAP == 2 and imgDimsAPDir == 1:
                dat[:, :, :cutFrom] = 0
            # set to label
            if imgDimsAP == 0:
                dat[cutFrom, np.nonzero(dat[cutFrom, :, :])[0], np.nonzero(dat[cutFrom, :, :])[1]] = labelTail
            elif imgDimsAP == 1:
                dat[np.nonzero(dat[:, cutFrom, :])[0], cutFrom, np.nonzero(dat[:, cutFrom, :])[1]] = labelTail
            elif imgDimsAP == 2:
                dat[np.nonzero(dat[:, :, cutFrom])[0], np.nonzero(dat[:, :, cutFrom])[1], cutFrom] = labelTail

        # create image
        out = nb.MGHImage(dat, img.affine, header=img.header)

        # save
        nb.save(out, os.path.join(params.OUTDIR, "labels", params.HEMI + ".automask_image.mgz"))

        # update params
        params.FILENAME = os.path.join(params.OUTDIR, "labels", params.HEMI + ".automask_image.mgz")

    # return

    return params


def createLabels(params):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Create labels")
    print()

    # create a mask
    cmd = (
        os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_binarize")
        + " "
        + "--i "
        + params.FILENAME
        + " "
        + "--match "
        + " ".join([str(x) for x in params.HSFLIST])
        + " "
        + "--o "
        + os.path.join(params.OUTDIR, "labels", params.HEMI + ".initial_labels.mgz")
    )

    print(cmd)

    subprocess.run(cmd.split(), capture_output=True)

    # multiply original image with mask
    cmd = (
        os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "fscalc")
        + " "
        + params.FILENAME
        + " mul "
        + os.path.join(params.OUTDIR, "labels", params.HEMI + ".initial_labels.mgz")
        + " --o "
        + os.path.join(params.OUTDIR, "labels", params.HEMI + ".initial_labels.mgz")
    )

    print(cmd)

    subprocess.run(cmd.split(), capture_output=True)

    # update params
    params.FILENAME = os.path.join(params.OUTDIR, "labels", params.HEMI + ".initial_labels.mgz")

    #
    return params


def mergeMolecularLayer(params):
    """ """

    if params.internal.MERGE_MOLECULAR_LAYER is True:
        # message
        print()
        print("--------------------------------------------------------------------------------")
        print("Attaching the molecular layer")
        print()

        # settings
        offset = 0

        # load image
        im = nb.load(params.FILENAME)

        # get voxel data
        vx = im.get_fdata()

        # get all voxels that are not zero; sort by dimensions 1,2,3
        vxNZ = list(np.nonzero(vx))
        vxNZ.append(vx[np.nonzero(vx)])
        vxNZ = np.transpose(vxNZ)
        vxNZ = vxNZ[np.argsort(vxNZ[:, 0]), :]
        vxNZ = vxNZ[np.argsort(vxNZ[:, 1], kind="mergesort"), :]
        vxNZ = vxNZ[np.argsort(vxNZ[:, 2], kind="mergesort"), :]

        # get all ML voxels
        vxML = list(np.where((vx == 245) | (vx == 246) | (vx == 214)))
        vxML.append(vx[np.where((vx == 245) | (vx == 246) | (vx == 214))])
        vxML = np.transpose(vxML)
        vxML = vxML[np.argsort(vxML[:, 0]), :]
        vxML = vxML[np.argsort(vxML[:, 1], kind="mergesort"), :]
        vxML = vxML[np.argsort(vxML[:, 2], kind="mergesort"), :]

        #
        for i in range(np.shape(vxML)[0]):
            n = 1
            tmpML = np.transpose((np.empty([0]), np.empty([0]), np.empty([0])))
            while np.shape(tmpML)[0] == 0:
                # break if iterations are exceeded or if coordinate is below zero or equal / above image dims
                if (
                    n > 10
                    or (int((vxML[i, 0] - n)) < 0)
                    or (int((vxML[i, 0] + n)) > (vx.shape[0] - 1))
                    or (int((vxML[i, 1] - n)) < 0)
                    or (int((vxML[i, 1] + n)) > (vx.shape[1] - 1))
                    or (int((vxML[i, 2] - n)) < 0)
                    or (int((vxML[i, 2] + n)) > (vx.shape[2] - 1))
                ):
                    vxML[i, 3] = 0
                    break
                tmp = vx[
                    np.ix_(
                        range(int((vxML[i, 0] - n)), int((vxML[i, 0] + n) + 1)),
                        range(int((vxML[i, 1] - n)), int((vxML[i, 1] + n) + 1)),
                        range(int((vxML[i, 2] - n)), int((vxML[i, 2] + n) + 1)),
                    )
                ]
                tmpML = np.transpose(np.where((tmp != 0) & (tmp != 246) & (tmp != 245) & (tmp != 214)))
                if np.shape(tmpML)[0] != 0:
                    tmp0 = np.asarray(tmpML[np.where(np.sum(tmpML, axis=1) == min(np.sum(tmpML, axis=1)))[0], :])
                    vxML[i, 3] = offset + min(tmp[tmp0[:, 0], tmp0[:, 1], tmp0[:, 2]])
                    tmp0 = None
                else:
                    n = n + 1
                tmp = None
            tmpML = None

        # write back to vx and im
        for i in range(np.shape(vxML)[0]):
            vx[int(vxML[i, 0]), int(vxML[i, 1]), int(vxML[i, 2])] = vxML[i, 3]

        om = nb.MGHImage(dataobj=vx, affine=im.affine, header=im.header)
        nb.save(img=om, filename=os.path.join(params.OUTDIR, "labels", params.HEMI + ".ml_labels.mgz"))

        # update params
        params.FILENAME = os.path.join(params.OUTDIR, "labels", params.HEMI + ".ml_labels.mgz")

    # return
    return params


def copy_labels_to_main(params):
    """ """

    # copy to main directory

    shutil.copyfile(params.FILENAME, os.path.join(params.OUTDIR, params.HEMI + ".labels.mgz"))

    # update params

    params.FILENAME = os.path.join(params.OUTDIR, params.HEMI + ".labels.mgz")

    # return

    return params



"""
This module provides functions for basic image processing operations (conversion, cropping, upsampling)

"""

import logging
import os
import shutil
import subprocess

import nibabel as nb
from nilearn import image as nli

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def convertFormat(params):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Convert to mgz and copy to output directory")
    print()

    # convert and copy

    nb.save(nb.load(params.FILENAME), os.path.join(params.OUTDIR, "image", params.HEMI + ".orig.mgz"))

    # update params

    params.FILENAME = os.path.join(params.OUTDIR, "image", params.HEMI + ".orig.mgz")

    # return

    return params


def cropImage(params):
    """ """

    if params.internal.CROP is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Cropping")
        print()

        # crop

        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_mask")
            + " "
            + "-bb 5 "
            + params.FILENAME
            + " "
            + params.FILENAME
            + " "
            + os.path.join(params.OUTDIR, "image", params.HEMI + ".cropped.mgz")
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

        # update params

        params.FILENAME = os.path.join(params.OUTDIR, "image", params.HEMI + ".cropped.mgz")

    # return

    return params


def upsampleImage(params):
    """ """

    if params.internal.UPSAMPLE is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Upsampling")
        print()

        # upsample

        if any(x != 0 for x in params.internal.UPSAMPLE_SIZE):
            LOGGER.info("Upsampling with custom parameters ...")

            cmd = (
                os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_convert")
                + " "
                + " -ds "
                + str(params.internal.UPSAMPLE_SIZE[0])
                + " "
                + str(params.internal.UPSAMPLE_SIZE[1])
                + "  "
                + str(params.internal.UPSAMPLE_SIZE[2])
                + " "
                + " -rt nearest "
                + params.FILENAME
                + " "
                + os.path.join(params.OUTDIR, "image", params.HEMI + ".upsampled.mgz")
            )

            print(cmd)

            subprocess.run(cmd.split(), capture_output=True)

        else:
            #
            LOGGER.info("Upsampling to min voxel size ...")

            # get image and info
            img = nb.load(params.FILENAME)
            vxsz = img.header["delta"]
            affn = img.affine

            # compute new affine
            target_affn = affn.copy()
            target_affn[0:4, 0] *= min(vxsz) / vxsz[0]
            target_affn[0:4, 1] *= min(vxsz) / vxsz[1]
            target_affn[0:4, 2] *= min(vxsz) / vxsz[2]

            # resample
            img_int = nli.resample_img(img, target_affine=target_affn, interpolation="nearest")

            # save
            nb.save(img_int, os.path.join(params.OUTDIR, "image", params.HEMI + ".upsampled.mgz"))

        # update params
        params.FILENAME = os.path.join(params.OUTDIR, "image", params.HEMI + ".upsampled.mgz")

    # return

    return params


def copy_image_to_main(params):
    """ """

    # copy to main directory

    shutil.copyfile(params.FILENAME, os.path.join(params.OUTDIR, params.HEMI + ".image.mgz"))

    # update params

    params.FILENAME = os.path.join(params.OUTDIR, params.HEMI + ".image.mgz")

    # return

    return params



"""
This module creates supplementary files, primarily for visualization.

"""

import logging
import os

import nibabel as nb
import numpy as np
from lapy import TriaMesh

from .get_levelsets import levelsetsTria

# ------------------------------------------------------------------------------
# MAIN FUNCTION
# ------------------------------------------------------------------------------


def createSupplementaryFiles(params):
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Creating supplementary files")
    print()

    # -------------------------------------------------------
    # get params

    LUT = params.LUT
    HEMI = params.HEMI
    OUT_DIR = os.path.join(params.OUTDIR, "thickness")

    # --------------------------------------------------------------------------
    # create overlay with subfield boundaries

    # list of boundary values:

    if LUT == "freesurfer":
        # get labels

        hsfBnd = nb.load(os.path.join(OUT_DIR, HEMI + ".mid-surface.hsf.mgh"))
        hsfBnd = np.array(hsfBnd.get_fdata()).flatten()

        # recode to consecutive values

        hsfBnd[hsfBnd == 234] = 1
        hsfBnd[hsfBnd == 236] = 2
        hsfBnd[hsfBnd == 238] = 3
        hsfBnd[hsfBnd == 240] = 4

        lstBnd = [
            1.5,
            2.5,
            3.5,
        ]  # note that this implicitly also accounts for 246 labels if present

    elif LUT == "ashs-penn_abc_3t_t2":
        # get labels

        hsfBnd = nb.load(os.path.join(OUT_DIR, HEMI + ".mid-surface.hsf.mgh"))
        hsfBnd = np.array(hsfBnd.get_fdata()).flatten()

        # recode to consecutive values

        hsfBnd[hsfBnd == 8] = 11
        hsfBnd[hsfBnd == 1] = 12
        hsfBnd[hsfBnd == 2] = 13
        hsfBnd[hsfBnd == 4] = 14

        lstBnd = [11.5, 12.5, 13.5]

    elif LUT == "ashs-umcutrecht_7t":
        # get labels

        hsfBnd = nb.load(os.path.join(OUT_DIR, HEMI + ".mid-surface.hsf.mgh"))
        hsfBnd = np.array(hsfBnd.get_fdata()).flatten()

        # recode to consecutive values

        hsfBnd[hsfBnd == 2] = 21
        hsfBnd[hsfBnd == 3] = 22
        hsfBnd[hsfBnd == 4] = 23
        hsfBnd[hsfBnd == 6] = 24

        lstBnd = [21.5, 22.5, 23.5]

    else:
        logging.info(
            "Creation of subfield boundary overlays is only supported for the freesurfer and ashs look-up tables."
        )
        raise ValueError()

    # get mid-surface

    triaMidRm = TriaMesh.read_vtk(os.path.join(OUT_DIR, HEMI + ".mid-surface.vtk"))

    #

    vBnd = np.empty((0, 3))
    tBnd = np.empty((0, 3))

    for i in lstBnd:
        lvlBnd = levelsetsTria(triaMidRm.v, triaMidRm.t, hsfBnd, i)
        if len(lvlBnd[0][0]) > 0:
            vBnd_lvl = np.array(lvlBnd[0][0])
            tBnd_lvl = np.array(lvlBnd[1][0])[:, [0, 1, 1]] - 1
            tBnd = np.concatenate((tBnd, tBnd_lvl + len(vBnd)), axis=0)
            vBnd = np.concatenate((vBnd, vBnd_lvl), axis=0)

    # output

    if len(vBnd) > 0:
        TriaMesh(v=vBnd, t=tBnd.astype("int")).write_vtk(
            filename=os.path.join(OUT_DIR, HEMI + ".mid-surface.hsf-bnd.vtk")
        )
    else:
        logging.info("Could not create boundary overlays.")



"""

"""

# ------------------------------------------------------------------------------
# get_version()


def get_version():
    from importlib import metadata

    try:
        # requires existing installation
        version = metadata.version("hipsta")
    except Exception:
        # fall-back if package is not installed, but run directly
        version = "unknown"

    return version



"""
This module provides documentation for the hippocampal shape and 
thickness analysis package.

"""

# this is a backport to python 3.8; for newer versions (>=3.10) use built-in
# importlib.resources
from importlib_resources import files


def get_help_text():
    txt = files("hipsta.doc").joinpath("DOCUMENTATION.md").read_text()

    return txt



"""

"""

import logging
import os
import sys
import tempfile
import time

from .version import get_version

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def setup_logging(args):
    """
    Set up logging
    """

    # check if output directory exists or can be created
    if os.path.isdir(args.outputdir):
        print("Found output directory " + args.outputdir)
    else:
        try:
            os.mkdir(args.outputdir)
        except OSError as e:
            print("Cannot create output directory " + args.outputdir)
            print("Reason: " + str(e))
            raise

    # check if logfile can be written in output directory
    try:
        testfile = tempfile.TemporaryFile(dir=args.outputdir)
        testfile.close()
    except OSError as e:
        print(args.outputdir + " not writeable")
        print("Reason: " + str(e))
        raise

    # setup logging
    logfile = os.path.join(args.outputdir, "logfile.txt")
    logfile_format = "[%(levelname)s: %(filename)s] %(message)s"
    logfile_handlers_stream = logging.StreamHandler(sys.stdout)
    logfile_handlers_stream.setFormatter(logging.Formatter(logfile_format))
    logfile_handlers_stream.setLevel(logging.INFO)
    logfile_handlers_file = logging.FileHandler(filename=logfile, mode="w")
    logfile_handlers_file.setFormatter(logging.Formatter(logfile_format))
    logfile_handlers_file.setLevel(logging.INFO)

    logging.basicConfig(
        level=logging.INFO, format=logfile_format, handlers=[logfile_handlers_file, logfile_handlers_stream]
    )

    # check if file handlers is part of the root logger; usually this will be
    # the case after the basicConfig() command is executed. However, if logging
    # configuration is done earlier, e.g. via pytest, then basicConfig() will
    # have no effect, and we need to do it manually as below:
    ROOT_LOGGER = logging.getLogger("root")
    if logfile_handlers_file not in ROOT_LOGGER.handlers:
        ROOT_LOGGER.addHandler(logfile_handlers_file)

    # initial messages (do not use logging earlier, since this will do the
    # default setup, and subsequent custom setup will have no effects)
    LOGGER.info("Starting logging ...")
    LOGGER.info("Logfile: %s", logfile)
    LOGGER.info("Version: %s", get_version())
    LOGGER.info("Date: %s", time.strftime("%d/%m/%Y %H:%M:%S"))

    # log args
    LOGGER.info("Command: " + " ".join(sys.argv))



"""
This module contains a set of auxiliary functions for mesh processing,
specifically:

- levelsetsTria(v, t, p, levelsets)
- levelsetsTetra(v, t, p, levelsets)

"""

# ------------------------------------------------------------------------------
#


def levelsetsTria(v, t, p, levelsets):
    import numpy as np
    from scipy.sparse import lil_matrix

    vLVL = list()
    lLVL = list()
    iLVL = list()

    levelsets = np.array(levelsets, ndmin=2)

    for l_i in range(len(levelsets)):
        A = lil_matrix((np.shape(v)[0], np.shape(v)[0]))

        lvl = levelsets[l_i]

        nlvl = p[t] > lvl

        n = np.where(np.logical_or(np.sum(nlvl, axis=1) == 1, np.sum(nlvl, axis=1) == 2))[0]

        # interpolate points

        ti = list()
        vi = list()

        for i in range(len(n)):
            # which are the outlying points in the current tria?
            oi = np.where(nlvl[n[i], :])[0]

            #  convert 2 --> 1
            if len(oi) == 2:
                oi = np.setdiff1d((0, 1, 2), oi)

            # find the two non - outyling points
            oix = np.setdiff1d((0, 1, 2), oi)

            #
            oi = oi.item()

            # check if we have interpolated for one or both of these points before

            if np.count_nonzero(A[t[n[i], oi], t[n[i], oix[0]]]) == 0:
                # compute difference vectors between outlying point and other points

                d10 = v[t[n[i], oix[0]], :] - v[t[n[i], oi], :]

                # compute differences of all points to lvl to get interpolation factors

                s10 = (lvl - p[t[n[i], oi]]) / (p[t[n[i], oix[0]]] - p[t[n[i], oi]])

                # compute new points

                v10 = s10 * d10 + v[t[n[i], oi], :]

                # update vi and index(order matters)
                vi.append(v10.tolist())

                ti10 = len(vi)

                # store between which two points we are interpolating (to avoid having duplicate points)

                A[t[n[i], oi], t[n[i], oix[0]]] = ti10
                A[t[n[i], oix[0]], t[n[i], oi]] = ti10

            else:
                ti10 = int(A[t[n[i], oi], t[n[i], oix[0]]])

            # essentially the same as above, just for oix[1]

            if np.count_nonzero(A[t[n[i], oi], t[n[i], oix[1]]]) == 0:
                d20 = v[t[n[i], oix[1]], :] - v[t[n[i], oi], :]

                s20 = (lvl - p[t[n[i], oi]]) / (p[t[n[i], oix[1]]] - p[t[n[i], oi]])

                v20 = s20 * d20 + v[t[n[i], oi], :]

                # update vi and index(order matters)
                vi.append(v20.tolist())

                ti20 = len(vi)

                A[t[n[i], oi], t[n[i], oix[1]]] = ti20
                A[t[n[i], oix[1]], t[n[i], oi]] = ti20

            else:
                ti20 = int(A[t[n[i], oi], t[n[i], oix[1]]])

            # store new indices
            ti.append((ti10, ti20))

        # store

        vLVL.append(vi)
        lLVL.append(ti)
        iLVL.append(n)

    return vLVL, lLVL, iLVL


# ------------------------------------------------------------------------------
# levelsetsTetra


def levelsetsTetra(v, t, p, levelsets):
    import numpy as np
    from scipy.sparse import lil_matrix

    vLVL = list()
    tLVL = list()
    iLVL = list()
    jLVL = list()
    oLVL = list()

    levelsets = np.array(levelsets, ndmin=2)

    for l_i in range(len(levelsets[0])):
        # matrix to store previous interpolation

        A = lil_matrix((np.shape(v)[0], np.shape(v)[0]))

        # current levelset

        lvl = levelsets[0][l_i]

        # for each tetra, which of its four points are outlying? nlvl is yes/no, n contains indices

        nlvl = p[t] > lvl

        n = np.where(
            np.logical_or(
                np.sum(nlvl, axis=1) == 1,
                np.logical_or(np.sum(nlvl, axis=1) == 2, np.sum(nlvl, axis=1) == 3),
            )
        )[0]

        # interpolate points

        ti = list()
        vi = list()
        m = list()
        o = list()

        for i in range(len(n)):
            # which are the outlying points in the current tetra?

            oi = np.where(nlvl[n[i], :])[0]

            # store oi prior to converting 3 --> 1

            o.append(oi)

            #  convert 3 --> 1

            if len(oi) == 3:
                oi = np.setdiff1d((0, 1, 2, 3), oi)

            # interpolate

            if len(oi) == 1:
                # find the three non - outyling points

                oix = np.setdiff1d((0, 1, 2, 3), oi)

                # check if we have interpolated for one or more of these points before

                if np.sum(np.count_nonzero(A[t[n[i], oi.item()], t[n[i], oix[0]]])) == 0:
                    # compute difference vectors between outlying point and other points

                    d10 = v[t[n[i], oix[0]], :] - v[t[n[i], oi], :]

                    # compute differences of all points to lvl to get interpolation factors

                    s10 = (lvl - p[t[n[i], oi]]) / (p[t[n[i], oix[0]]] - p[t[n[i], oi]])

                    # compute new points

                    v10 = s10 * d10 + v[t[n[i], oi], :]

                    # update vi and index(order matters)
                    vi.append(v10.tolist()[0])  # OK [a][0] -> l

                    ti10 = len(vi)

                    # store between which two points we are interpolating (to avoid having duplicate points)

                    A[t[n[i], oi.item()], t[n[i], oix[0]]] = ti10
                    A[t[n[i], oix[0]], t[n[i], oi.item()]] = ti10

                else:
                    ti10 = int(A[t[n[i], oi.item()], t[n[i], oix[0]]])

                # essentially the same as above, just for oix[1]

                if np.sum(np.count_nonzero(A[t[n[i], oi.item()], t[n[i], oix[1]]])) == 0:
                    d20 = v[t[n[i], oix[1]], :] - v[t[n[i], oi], :]

                    s20 = (lvl - p[t[n[i], oi]]) / (p[t[n[i], oix[1]]] - p[t[n[i], oi]])

                    v20 = s20 * d20 + v[t[n[i], oi], :]

                    # update vi and index(order matters)
                    vi.append(v20.tolist()[0])  # OK [a][0] -> l

                    ti20 = len(vi)

                    A[t[n[i], oi.item()], t[n[i], oix[1]]] = ti20
                    A[t[n[i], oix[1]], t[n[i], oi.item()]] = ti20

                else:
                    ti20 = int(A[t[n[i], oi.item()], t[n[i], oix[1]]])

                # essentially the same as above, just for oix[2]

                if np.sum(np.count_nonzero(A[t[n[i], oi.item()], t[n[i], oix[2]]])) == 0:
                    d30 = v[t[n[i], oix[2]], :] - v[t[n[i], oi], :]

                    s30 = (lvl - p[t[n[i], oi]]) / (p[t[n[i], oix[2]]] - p[t[n[i], oi]])

                    v30 = s30 * d30 + v[t[n[i], oi], :]

                    # update vi and index(order matters)
                    vi.append(v30.tolist()[0])  # OK [a][0] -> l

                    ti30 = len(vi)

                    A[t[n[i], oi.item()], t[n[i], oix[2]]] = ti30
                    A[t[n[i], oix[2]], t[n[i], oi.item()]] = ti30

                else:
                    ti30 = int(A[t[n[i], oi.item()], t[n[i], oix[2]]])

                # store new indices
                ti.append((ti10, ti20, ti30))  # OK

                m.append(n[i])  # OK

            elif len(oi) == 2:
                # find the two non-outyling points

                oix = np.setdiff1d((0, 1, 2, 3), oi)

                # check if we have interpolated for one or more of these points
                # before

                if np.sum(np.count_nonzero(A[t[n[i], oi[0]], t[n[i], oix[0]]])) == 0:
                    # compute difference vectors between outlying point and other points

                    d20 = v[t[n[i], oix[0]], :] - v[t[n[i], oi[0]], :]

                    # compute differences of all points to lvl to get interpolation
                    # factors

                    s20 = (lvl - p[t[n[i], oi[0]]]) / (p[t[n[i], oix[0]]] - p[t[n[i], oi[0]]])

                    # compute new points

                    v20 = s20 * d20 + v[t[n[i], oi[0]], :]

                    # update vi and index (order matters)
                    vi.append(v20.tolist())  ### a->l OK

                    ti20 = len(vi)

                    # store between which two points we are interpolating (to
                    # avoid having duplicate points)

                    A[t[n[i], oi[0]], t[n[i], oix[0]]] = ti20
                    A[t[n[i], oix[0]], t[n[i], oi[0]]] = ti20

                else:
                    ti20 = int(A[t[n[i], oi[0]], t[n[i], oix[0]]])

                # same as above for oi[0] and oix[1]

                if np.sum(np.count_nonzero(A[t[n[i], oi[0]], t[n[i], oix[1]]])) == 0:
                    # compute difference vectors between outlying point and other points

                    d30 = v[t[n[i], oix[1]], :] - v[t[n[i], oi[0]], :]

                    # compute differences of all points to lvl to get interpolation
                    # factors

                    s30 = (lvl - p[t[n[i], oi[0]]]) / (p[t[n[i], oix[1]]] - p[t[n[i], oi[0]]])

                    # compute new points

                    v30 = s30 * d30 + v[t[n[i], oi[0]], :]

                    # update vi and index (order matters)
                    vi.append(v30.tolist())  # a->l OK

                    ti30 = len(vi)

                    # store between which two points we are interpolating (to
                    # avoid having duplicate points)

                    A[t[n[i], oi[0]], t[n[i], oix[1]]] = ti30
                    A[t[n[i], oix[1]], t[n[i], oi[0]]] = ti30

                else:
                    ti30 = int(A[t[n[i], oi[0]], t[n[i], oix[1]]])

                # same as above for oi[1] and oix[0]

                if np.sum(np.count_nonzero(A[t[n[i], oi[1]], t[n[i], oix[0]]])) == 0:
                    # compute difference vectors between outlying point and other points

                    d21 = v[t[n[i], oix[0]], :] - v[t[n[i], oi[1]], :]

                    # compute differences of all points to lvl to get interpolation
                    # factors

                    s21 = (lvl - p[t[n[i], oi[1]]]) / (p[t[n[i], oix[0]]] - p[t[n[i], oi[1]]])

                    # compute new points

                    v21 = s21 * d21 + v[t[n[i], oi[1]], :]

                    # update vi and index (order matters)
                    vi.append(v21.tolist())  # a->l OK

                    ti21 = len(vi)

                    # store between which two points we are interpolating (to
                    # avoid having duplicate points)

                    A[t[n[i], oi[1]], t[n[i], oix[0]]] = ti21
                    A[t[n[i], oix[0]], t[n[i], oi[1]]] = ti21

                else:
                    ti21 = int(A[t[n[i], oi[1]], t[n[i], oix[0]]])

                # same as above for oi[1] and oix[1]

                if np.sum(np.count_nonzero(A[t[n[i], oi[1]], t[n[i], oix[1]]])) == 0:
                    # compute difference vectors between outlying point and other points

                    d31 = v[t[n[i], oix[1]], :] - v[t[n[i], oi[1]], :]

                    # compute differences of all points to lvl to get interpolation
                    # factors

                    s31 = (lvl - p[t[n[i], oi[1]]]) / (p[t[n[i], oix[1]]] - p[t[n[i], oi[1]]])

                    # compute new points

                    v31 = s31 * d31 + v[t[n[i], oi[1]], :]

                    # update vi and index (order matters)
                    vi.append(v31.tolist())  # a->l OK

                    ti31 = len(vi)

                    # store between which two points we are interpolating (to
                    # avoid having duplicate points)

                    A[t[n[i], oi[1]], t[n[i], oix[1]]] = ti31
                    A[t[n[i], oix[1]], t[n[i], oi[1]]] = ti31

                else:
                    ti31 = int(A[t[n[i], oi[1]], t[n[i], oix[1]]])

                # store new indices

                ti.append((ti20, ti21, ti30))
                ti.append((ti21, ti30, ti31))

                m.append(n[i])
                m.append(n[i])

        # store

        vLVL.append(np.array(vi))  # interpolated points
        tLVL.append(np.array(ti))  # triangles of interpolated points (1-based indices)
        iLVL.append(n)  # indices of tetras with 1-3 outlying points
        jLVL.append(m)  # for each interpolated point, store its tetra
        oLVL.append(o)  # outlying points in the current tetra

    return vLVL, tLVL, iLVL, jLVL, oLVL
